#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2023, Intel Corporation
#

import os

import testframework as t


class PMEM2_MAP(t.Test):
    test_type = t.Short
    filesize = 16 * t.MiB
    with_size = True

    def run(self, ctx):
        filepath = ctx.create_holey_file(self.filesize, 'testfile',)
        if self.with_size:
            filesize = os.stat(filepath).st_size
            ctx.exec('pmem2_map', self.test_case, filepath, filesize)
        else:
            ctx.exec('pmem2_map', self.test_case, filepath)


class PMEM2_MAP_NO_FILE(t.Test):
    test_type = t.Short

    def run(self, ctx):
        ctx.exec('pmem2_map', self.test_case)


@t.require_devdax(t.DevDax('devdax1'))
class PMEM2_MAP_DEVDAX(t.Test):
    test_type = t.Short
    with_size = True

    def run(self, ctx):
        dd = ctx.devdaxes.devdax1
        if self.with_size:
            ctx.exec('pmem2_map', self.test_case, dd.path, dd.size)
        else:
            ctx.exec('pmem2_map', self.test_case, dd.path)


# XXX disable the test for `memcheck'
# until https://github.com/pmem/pmdk/issues/5600 is fixed.
@t.require_valgrind_disabled('memcheck')
class TEST0(PMEM2_MAP):
    """map a O_RDWR file"""
    test_case = "test_map_rdrw_file"
    with_size = False


# XXX disable the test for `memcheck'
# until https://github.com/pmem/pmdk/issues/5600 is fixed.
@t.require_valgrind_disabled('memcheck')
class TEST1(PMEM2_MAP_DEVDAX):
    """DevDax map a O_RDWR file"""
    test_case = "test_map_rdrw_file"
    with_size = False


class TEST2(PMEM2_MAP):
    """map a O_RDONLY file"""
    test_case = "test_map_rdonly_file"
    with_size = False


class TEST3(PMEM2_MAP_DEVDAX):
    """DevDax map a O_RDONLY file"""
    test_case = "test_map_rdonly_file"
    with_size = False


# XXX disable the test for `memcheck'
# until https://github.com/pmem/pmdk/issues/5600 is fixed.
@t.require_valgrind_disabled('memcheck')
class TEST4(PMEM2_MAP):
    """map valid memory ranges"""
    test_case = "test_map_valid_ranges"


# XXX disable the test for `memcheck'
# until https://github.com/pmem/pmdk/issues/5600 is fixed.
@t.require_valgrind_disabled('memcheck')
class TEST5(PMEM2_MAP_DEVDAX):
    """DevDax map valid memory ranges"""
    test_case = "test_map_valid_ranges"


class TEST6(PMEM2_MAP):
    """map invalid memory ranges"""
    test_case = "test_map_invalid_ranges"


class TEST7(PMEM2_MAP_DEVDAX):
    """DevDax map invalid memory ranges"""
    test_case = "test_map_invalid_ranges"


class TEST8(PMEM2_MAP):
    """map using invalid alignment in the offset"""
    test_case = "test_map_invalid_alignment"


class TEST9(PMEM2_MAP_DEVDAX):
    """DevDax map using invalid alignment in the offset"""
    test_case = "test_map_invalid_alignment"


class TEST10(PMEM2_MAP):
    """map using a invalid file descriptor"""
    test_case = "test_map_invalid_fd"


# XXX disable the test for `memcheck'
# until https://github.com/pmem/pmdk/issues/5600 is fixed.
@t.require_valgrind_disabled('memcheck')
class TEST11(PMEM2_MAP):
    """unmap valid pmem2 mapping"""
    test_case = "test_unmap_valid"


# XXX disable the test for `memcheck'
# until https://github.com/pmem/pmdk/issues/5600 is fixed.
@t.require_valgrind_disabled('memcheck')
class TEST12(PMEM2_MAP_DEVDAX):
    """DevDax unmap valid pmem2 mapping"""
    test_case = "test_unmap_valid"


# XXX disable the test for `memcheck'
# until https://github.com/pmem/pmdk/issues/5600 is fixed.
@t.require_valgrind_disabled('memcheck')
class TEST13(PMEM2_MAP):
    """unmap a pmem2 mapping with an invalid length"""
    test_case = "test_unmap_zero_length"


# XXX disable the test for `memcheck'
# until https://github.com/pmem/pmdk/issues/5600 is fixed.
@t.require_valgrind_disabled('memcheck')
class TEST14(PMEM2_MAP_DEVDAX):
    """DevDax unmap a pmem2 mapping with an invalid length"""
    test_case = "test_unmap_zero_length"


# UnmapViewOfFile does not care about the address alignment
# XXX disable the test for `memcheck'
# until https://github.com/pmem/pmdk/issues/5600 is fixed.
@t.require_valgrind_disabled('memcheck')
class TEST15(PMEM2_MAP):
    """unmap a pmem2 mapping with an unaligned address"""
    test_case = "test_unmap_unaligned_addr"


# XXX disable the test for `memcheck'
# until https://github.com/pmem/pmdk/issues/5600 is fixed.
@t.require_valgrind_disabled('memcheck')
class TEST16(PMEM2_MAP_DEVDAX):
    """DevDax unmap a pmem2 mapping with an unaligned address"""
    test_case = "test_unmap_unaligned_addr"


class TEST18(PMEM2_MAP_NO_FILE):
    """test for pmem2_map_get_address"""
    test_case = "test_map_get_address"


class TEST19(PMEM2_MAP_NO_FILE):
    """test for pmem2_map_get_size"""
    test_case = "test_map_get_size"


class TEST20(PMEM2_MAP_NO_FILE):
    """simply get the previously stored value of granularity"""
    test_case = "test_get_granularity_simple"


class TEST21(PMEM2_MAP):
    """map a file of length which is not page-aligned"""
    test_case = "test_map_unaligned_length"
    filesize = 3 * t.KiB


class TEST22(PMEM2_MAP):
    """map a file which size is not aligned"""
    test_case = "test_map_larger_than_unaligned_file_size"
    filesize = 16 * t.MiB - 1


class TEST23(PMEM2_MAP):
    """
    map a file with zero size, do not provide length
    to pmem2_map config
    """
    test_case = "test_map_zero_file_size"
    filesize = 0


# XXX disable the test for `memcheck'
# until https://github.com/pmem/pmdk/issues/5600 is fixed.
@t.require_valgrind_disabled('memcheck')
class TEST24(PMEM2_MAP):
    """
    map a file with PMEM2_SHARED sharing, changes in the mapping are visible
    in another mapping
    """
    test_case = "test_map_sharing_shared"
    with_size = False


# XXX disable the test for `memcheck'
# until https://github.com/pmem/pmdk/issues/5600 is fixed.
@t.require_valgrind_disabled('memcheck')
class TEST25(PMEM2_MAP):
    """
    map a file with PMEM2_PRIVATE sharing, changes in the mapping are not
    visible in another mapping
    """
    test_case = "test_map_sharing_private"
    with_size = False


# XXX disable the test for `memcheck'
# until https://github.com/pmem/pmdk/issues/5600 is fixed.
@t.require_valgrind_disabled('memcheck')
class TEST26(PMEM2_MAP):
    """
    map a file with PMEM2_PRIVATE sharing, changes in the mapping are not
    visible in another mapping, fd is reopened before each mapping
    """
    test_case = "test_map_sharing_private_with_reopened_fd"
    with_size = False


# XXX disable the test for `memcheck'
# until https://github.com/pmem/pmdk/issues/5600 is fixed.
@t.require_valgrind_disabled('memcheck')
class TEST27(PMEM2_MAP):
    """
    map O_RDONLY file with PMEM2_PRIVATE sharing
    """
    test_case = "test_map_sharing_private_rdonly_file"
    with_size = False


class TEST28(PMEM2_MAP_DEVDAX):
    """DevDax file with PMEM2_PRIVATE sharing"""
    test_case = "test_map_sharing_private_devdax"
    with_size = False


@t.require_architectures('x86_64')
# XXX disable the test for `memcheck'
# until https://github.com/pmem/pmdk/issues/5600 is fixed.
@t.require_valgrind_disabled('memcheck')
class TEST29(PMEM2_MAP):
    """map alignment test for huge pages"""
    test_case = "test_map_huge_alignment"
    filesize = 16 * t.MiB


@t.require_architectures('x86_64')
# XXX disable the test for `memcheck'
# until https://github.com/pmem/pmdk/issues/5600 is fixed.
@t.require_valgrind_disabled('memcheck')
class TEST30(PMEM2_MAP):
    """map alignment test for small pages"""
    test_case = "test_map_huge_alignment"
    filesize = 16 * t.KiB



#!usr/bin/env python3
#
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2018-2023, Intel Corporation

"""
This module includes tests for PMDK packages installation.
Tests check:
-compatibility of the version of installed  packages from PMDK library with
 the current version of PMDK library,
-if all packages from PMDK library are installed.
Required arguments:
-r <PMDK_path>    the PMDK library root path.
"""

from os import listdir, path, linesep, walk
from subprocess import check_output
import distro
import unittest
import xmlrunner
import sys
import re

NO_PKG_CONFIGS = ('pmdk', 'pmempool', 'daxio', 'pmreorder', 'rpmemd')
PMDK_TOOLS = ('pmempool', 'daxio', 'pmreorder', 'rpmemd')
PMDK_VERSION = ''
SYSTEM_ARCHITECTURE = ''
PMDK_PATH = ''


def get_package_version_and_system_architecture():
    """
    Returns packages version and system architecture from names of directories
    from pkg directory.
    """
    global PMDK_VERSION
    global SYSTEM_ARCHITECTURE
    os_distro = distro.id()
    if os_distro != 'ubuntu':
        pkg_directory = path.join(PMDK_PATH, 'rpm')
        for elem in listdir(pkg_directory):
            if '.src.rpm' in elem:
                # looks for the version number of package in package name
                PMDK_VERSION = re.search(
                    r'[\s]*pmdk-([\S]+).src.rpm', elem).group(1)
            else:
                    SYSTEM_ARCHITECTURE = elem

    elif os_distro == 'ubuntu':
        pkg_directory = path.join(PMDK_PATH, 'dpkg')
        for elem in listdir(pkg_directory):
            if '.changes' in elem:
                # looks for the version number of package in package name
                PMDK_VERSION = re.search(r'pmdk_*(.+)_(.+).changes', elem).group(1)
                SYSTEM_ARCHITECTURE = re.search(r'pmdk_*(.+)_(.+).changes', elem).group(2)


def get_libraries_names(packages_path, split_param):
    """
    Returns names of elements, for which are installed packages from PMDK
    library.
    """
    libraries_names = [re.split(split_param, elem)[0] for elem in listdir(packages_path)
                           if PMDK_VERSION in elem]

    for i in range(len(libraries_names)):
        if libraries_names[i].endswith('1'):
            libraries_names[i] = libraries_names[i].replace('1', '')

    return set(libraries_names)


def get_not_installed_packages(packages_path, so_path, split_param):
    """
    Returns names of packages from PMDK library, which are not installed.
    """
    def is_installed(elem):
        return elem in PMDK_TOOLS and elem in listdir('/usr/bin/') or\
            elem == "pmdk" or elem + '.so' in listdir(so_path)

    elements = get_libraries_names(packages_path, split_param)
    not_installed_packages = []
    for elem in elements:
        if not is_installed(elem):
            not_installed_packages.append(elem)
    return not_installed_packages


def get_incompatible_packages(packages_path, pkgconfig_directory, split_param):
    """
    Returns names of packages from PMDK library, which are not compatible
    with the current version of PMDK library.
    """
    incompatibe_packages = []
    libraries = get_libraries_names(packages_path, split_param) - set(NO_PKG_CONFIGS)
    for library in libraries:
        with open(pkgconfig_directory + library + '.pc') as f:
            out = f.readlines()
        for line in out:
            if 'version=' in line:
                version = line.split('=')[1].strip(linesep)
        if not version in PMDK_VERSION.replace('~', '-'):
            incompatibe_packages.append(library)
    return incompatibe_packages


class TestBuildPackages(unittest.TestCase):

    def test_compatibility_of_version_of_installed_packages(self):
        """
        Checks if the version of installed packages is correct.
        """
        incompatible_packages = get_incompatible_packages(packages_path, pkgconfig_directory, split_param)
        error_msg = linesep + 'List of incompatible packages: '
        for package in incompatible_packages:
            error_msg += linesep + package
        self.assertFalse(incompatible_packages, error_msg)

    def test_correctness_of_installed_packages(self):
        """
        Checks if all packages from PMDK library are installed.
        """
        not_installed_packages = get_not_installed_packages(packages_path, so_path, split_param)
        error_msg = linesep + 'List of not installed packages: '
        for package in not_installed_packages:
            error_msg += linesep + package
        self.assertFalse(not_installed_packages, error_msg)


def parse_argument(argument_option):
    """
    Parses an option from the command line.
    """
    index = sys.argv.index(argument_option)
    try:
        argument = sys.argv[index+1]
    except IndexError:
        print('ERROR: Invalid argument!')
        print(__doc__)
        print(unittest.main.__doc__)
    else:
        sys.argv.pop(index)
        sys.argv.pop(index)
        return argument


if __name__ == '__main__':

    if '-h' in sys.argv or '--help' in sys.argv:
        print(__doc__)
        unittest.main()
    elif '-r' in sys.argv:
        PMDK_PATH = parse_argument('-r')
        get_package_version_and_system_architecture()
        os_distro = distro.id()
        if os_distro != 'ubuntu':
            packages_path = path.join(PMDK_PATH, 'rpm', SYSTEM_ARCHITECTURE)
            pkgconfig_directory = '/usr/lib64/pkgconfig/'
            so_path = '/usr/lib64/'
            split_param = '-'
        elif os_distro == 'ubuntu':
            packages_path = path.join(PMDK_PATH, 'dpkg')
            pkgconfig_directory = '/lib/x86_64-linux-gnu/pkgconfig/'
            so_path = '/lib/x86_64-linux-gnu/'
            split_param = '[-_]'

        if PMDK_VERSION == '' or SYSTEM_ARCHITECTURE == '':
            sys.exit("FATAL ERROR: command 'make rpm/dpkg' was not done correctly")
        unittest.main(
                testRunner=xmlrunner.XMLTestRunner(output='test-reports'),
                # these make sure that some options that are not applicable
                # remain hidden from the help menu.
                failfast=False, buffer=False, catchbreak=False)
    else:
        print(__doc__)
        print(unittest.main.__doc__)



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2023, Intel Corporation
#


import testframework as t
from testframework import granularity as g


class Granularity(str):
    BYTE = '0'
    CACHE_LINE = '1'
    PAGE = '2'


class PMEM2_INTEGRATION(t.Test):
    test_type = t.Medium

    def run(self, ctx):
        filepath = ctx.create_holey_file(16 * t.MiB, 'testfile')
        ctx.exec('pmem2_integration', self.test_case, filepath)


@t.require_devdax(t.DevDax('devdax'), max_size=1 * t.GiB)
class PMEM2_INTEGRATION_DEV_DAXES(t.Test):
    test_type = t.Medium

    def run(self, ctx):
        dd = ctx.devdaxes.devdax
        print(dd.path)
        ctx.exec('pmem2_integration', self.test_case, dd.path)


class PMEM2_GRANULARITY(t.Test):
    test_type = t.Medium
    test_case = 'test_granularity'
    available_granularity = None
    requested_granularity = None

    def run(self, ctx):
        filepath = ctx.create_holey_file(16 * t.MiB, 'testfile')
        ctx.exec('pmem2_integration', self.test_case, filepath,
                 self.available_granularity, self.requested_granularity)


class TEST0(PMEM2_INTEGRATION):
    """map twice using the same config"""
    test_case = "test_reuse_cfg"


class TEST1(PMEM2_INTEGRATION):
    """map using the same config with changed file descriptor"""
    test_case = "test_reuse_cfg_with_diff_fd"

    def run(self, ctx):
        filepath1 = ctx.create_holey_file(16 * t.MiB, 'testfile1')
        filepath2 = ctx.create_holey_file(16 * t.MiB, 'testfile2')
        ctx.exec('pmem2_integration', self.test_case, filepath1, filepath2)


@t.require_valgrind_enabled('pmemcheck')
class TEST2(PMEM2_INTEGRATION):
    """check if Valgrind registers data writing on pmem"""
    test_case = "test_register_pmem"


@t.require_valgrind_enabled('pmemcheck')
class TEST3(PMEM2_INTEGRATION_DEV_DAXES):
    """check if Valgrind registers data writing on DevDax"""
    test_case = "test_register_pmem"


class TEST4(PMEM2_INTEGRATION):
    """create multiple mappings with different offsets and lengths"""
    test_case = "test_use_misc_lens_and_offsets"

    def run(self, ctx):
        filepath = ctx.create_holey_file(1 * t.MiB, 'testfile1')
        ctx.exec('pmem2_integration', self.test_case, filepath)


@g.require_granularity(g.PAGE)
class TEST5(PMEM2_GRANULARITY):
    """test granularity with available page granularity and expected page
    granularity"""
    available_granularity = Granularity.PAGE
    requested_granularity = Granularity.PAGE


@g.require_granularity(g.PAGE)
class TEST6(PMEM2_GRANULARITY):
    """test granularity with available page granularity and expected cache
    line granularity"""
    available_granularity = Granularity.PAGE
    requested_granularity = Granularity.CACHE_LINE


@g.require_granularity(g.PAGE)
class TEST7(PMEM2_GRANULARITY):
    """test granularity with available page granularity and expected byte
    granularity"""
    available_granularity = Granularity.PAGE
    requested_granularity = Granularity.BYTE


@g.require_granularity(g.CACHELINE)
class TEST8(PMEM2_GRANULARITY):
    """test granularity with available cache line granularity and expected
    page granularity"""
    available_granularity = Granularity.CACHE_LINE
    requested_granularity = Granularity.PAGE


@g.require_granularity(g.CACHELINE)
class TEST9(PMEM2_GRANULARITY):
    """test granularity with available cache line granularity and expected
    cache line granularity"""
    available_granularity = Granularity.CACHE_LINE
    requested_granularity = Granularity.CACHE_LINE


@g.require_granularity(g.CACHELINE)
class TEST10(PMEM2_GRANULARITY):
    """test granularity with available cache line granularity and expected
    byte granularity"""
    available_granularity = Granularity.CACHE_LINE
    requested_granularity = Granularity.BYTE


@g.require_granularity(g.BYTE)
class TEST11(PMEM2_GRANULARITY):
    """test granularity with available byte granularity and expected page
    granularity"""
    available_granularity = Granularity.BYTE
    requested_granularity = Granularity.PAGE


@g.require_granularity(g.BYTE)
class TEST12(PMEM2_GRANULARITY):
    """test granularity with available byte granularity and expected cache
    line granularity"""
    available_granularity = Granularity.BYTE
    requested_granularity = Granularity.CACHE_LINE


@g.require_granularity(g.BYTE)
class TEST13(PMEM2_GRANULARITY):
    """test granularity with available byte granularity and expected byte
    granularity"""
    available_granularity = Granularity.BYTE
    requested_granularity = Granularity.BYTE


class TEST14(PMEM2_INTEGRATION):
    """test not aligned length"""
    test_case = "test_len_not_aligned"


class TEST15(PMEM2_INTEGRATION_DEV_DAXES):
    """test not aligned length on DevDax"""
    test_case = "test_len_not_aligned"


class TEST16(PMEM2_INTEGRATION):
    """test aligned length"""
    test_case = "test_len_aligned"


class TEST17(PMEM2_INTEGRATION_DEV_DAXES):
    """test aligned length on DevDax"""
    test_case = "test_len_aligned"


class TEST18(PMEM2_INTEGRATION):
    """test unaligned offset"""
    test_case = "test_offset_not_aligned"


class TEST19(PMEM2_INTEGRATION_DEV_DAXES):
    """test unaligned offset"""
    test_case = "test_offset_not_aligned"


class TEST20(PMEM2_INTEGRATION):
    """test unaligned offset"""
    test_case = "test_offset_aligned"


class TEST21(PMEM2_INTEGRATION_DEV_DAXES):
    """test unaligned offset"""
    test_case = "test_offset_aligned"


class TEST22(PMEM2_INTEGRATION):
    """
    map O_RDONLY file and test pmem2_[cpy|set|move]_fns with
    PMEM2_PRIVATE sharing
    """
    test_case = "test_mem_move_cpy_set_with_map_private"


class TEST23(PMEM2_INTEGRATION):
    """test valid case of pmem2_deep_sflush"""
    test_case = "test_deep_flush_valid"


class TEST24(PMEM2_INTEGRATION):
    """test deep flush with range out of map"""
    test_case = "test_deep_flush_e_range_behind"


class TEST25(PMEM2_INTEGRATION):
    """test deep flush with range out of map"""
    test_case = "test_deep_flush_e_range_before"


class TEST26(PMEM2_INTEGRATION):
    """test deep flush with part of map"""
    test_case = "test_deep_flush_slice"


class TEST27(PMEM2_INTEGRATION):
    """test deep flush with overlapping part"""
    test_case = "test_deep_flush_overlap"


class TEST28(PMEM2_INTEGRATION):
    """test for anonymous mappings"""
    test_case = "test_source_anon_private"


class TEST29(PMEM2_INTEGRATION):
    """test for anonymous mappings"""
    test_case = "test_source_anon_shared"


class TEST30(PMEM2_INTEGRATION):
    """test for anonymous mappings"""
    test_case = "test_source_anon_page"


class TEST31(PMEM2_INTEGRATION):
    """test for anonymous mappings"""
    test_case = "test_source_anon_too_small"


class TEST32(PMEM2_INTEGRATION):
    """test for anonymous mappings"""
    test_case = "test_source_anon_zero_len"


# The Valgrind pmemcheck tool does not have a notion of a deep flush.
# Whereas from the persistency point of view, the deep flush test cases are
# trivial so running them under pmemcheck produces little value.
# At the same time, they touch a whole provided device DAX space hence running
# them under pmemcheck is extensively time-consuming.
@t.require_valgrind_disabled('pmemcheck')
class PMEM2_INTEGRATION_DEV_DAXES_NO_PMEMCHECK(PMEM2_INTEGRATION_DEV_DAXES):
    pass


class TEST33(PMEM2_INTEGRATION_DEV_DAXES_NO_PMEMCHECK):
    """test valid case of pmem2_deep_sflush"""
    test_case = "test_deep_flush_valid"


class TEST34(PMEM2_INTEGRATION_DEV_DAXES_NO_PMEMCHECK):
    """test deep flush with range out of map"""
    test_case = "test_deep_flush_e_range_behind"


class TEST35(PMEM2_INTEGRATION_DEV_DAXES_NO_PMEMCHECK):
    """test deep flush with range out of map"""
    test_case = "test_deep_flush_e_range_before"


class TEST36(PMEM2_INTEGRATION_DEV_DAXES_NO_PMEMCHECK):
    """test deep flush with part of map"""
    test_case = "test_deep_flush_slice"


class TEST37(PMEM2_INTEGRATION_DEV_DAXES_NO_PMEMCHECK):
    """test deep flush with overlapping part"""
    test_case = "test_deep_flush_overlap"


class TEST38(PMEM2_INTEGRATION):
    """test for unaligned persists"""
    test_case = "test_unaligned_persist"


class TEST39(PMEM2_INTEGRATION):
    """compare normal map vs map_from_existing"""
    test_case = "test_map_from_existing"


class TES40(PMEM2_INTEGRATION):
    """test map_from_existing from pmem2_mapping"""
    test_case = "test_map_from_existing_map"


class TEST41(PMEM2_INTEGRATION_DEV_DAXES):
    """compare normal map vs map_from_existing on devdax"""
    test_case = "test_map_from_existing"



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2021-2023, Intel Corporation
#

import testframework as t
import re
import os
from testframework import granularity as g


@g.require_usc()
class Pmem2USC(t.Test):
    test_type = t.Short

    def run(self, ctx):
        old_log_level = ctx.env['UNITTEST_LOG_LEVEL']

        # writing to stdout file with UT_OUT requires log level >= 2
        if old_log_level is None or int(old_log_level) < 2:
            ctx.env['UNITTEST_LOG_LEVEL'] = '2'

        filepath = ctx.create_holey_file(1 * t.MiB, 'testfile')

        log_file = os.path.join(ctx.testdir, 'usc.log')
        ctx.exec('pmem2_usc', filepath, stdout_file=log_file)

        log_content = open(log_file).read()
        usc_line = re.findall("USC: .*", log_content, re.MULTILINE).pop()

        # get only the log content after USC log header
        usc_log_header = "USC: "
        usc = int(usc_line[len(usc_log_header):])
        usc_from_tool = ctx.usc.read(ctx.testdir)

        if usc_from_tool != usc:
            t.futils.fail("USC mismatch (tool: {}, pmem2: {})"
                          .format(usc_from_tool, usc))

        # restore previous log level
        ctx.env['UNITTEST_LOG_LEVEL'] = old_log_level


class TEST0(Pmem2USC):
    """check if pmem2 usc output is the same as ndctl/impctl output"""



# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2024, Intel Corporation

"""Base tests class and its functionalities.

_TestCase metaclass implemented in this module is
used to register test case classes.

Classes are registered and used as actual test cases only if they fulfill
the following requirements:
    - inherit (directly or indirectly) from BaseTest class
    - reside in TESTS.py file under pmdk/src/test/ directory
    - are named by TEST[int] (e.g. TEST0), where integer suffix is a
      test number (used for selecting test sequences).

BaseTest._execute() method requires special attention as a
general single test execution workflow.

While BaseTest class is the required test case base class, it consist only of
the minimal implementation to serve as the test framework handle.
In practice, it is the Test class (inheriting from BaseTest) that is
most commonly used as a test case direct base class. The Test class
implements operations typically used by most of the tests.

Test class needs to implement an abstract run(ctx) method, which stands
for a test body. setup(ctx) and clean(ctx) methods, which are a part
of test execution workflow are optional. Mind that the Test class already
provides some implementation of the setup() and clean() - it is
crucial to call them while implementing own setup() and clean() methods.

Tests for PMDK C code frequently call the Context exec() method
(passed as 'ctx' argument to run()), which is a utility for
setting up, running and handling the output of C test binaries.

Example TESTS.py file with a single test case may look like this:

$ cat pmdk/src/test/test_group/TESTS.py

!#../env.py

import testframework as t


class TEST0(t.Test):
    def run(self, ctx):
        ctx.exec('test_binary')

    def setup(self, ctx):
        # call Test setup
        super().setup(ctx)
        # TEST0 setup implementation goes here

    def clean(self, ctx):
        super().clean(ctx)
        # TEST0 clean implementation goes here

"""

import builtins
import subprocess as sp
import sys
import re
import os
from datetime import datetime
from os import path

from configurator import Configurator
from consts import LIBS_LIST, ROOTDIR
import granularity as g
import futils
import test_types
import shutil


#
# imported test cases are globally available as a 'builtins'
# module 'testcases' variable, which is initialized at first basetest
# module import as an empty list
#
if not hasattr(builtins, 'testcases'):
    builtins.testcases = []


def get_testcases():
    """"Get list of testcases imported from src/test tree"""
    return builtins.testcases


def _test_string_repr(cls):
    """
    Implementation of __str__ method for the test class. Needs to be available
    both for initialized (as a BaseTest instance method)
    as well as uninitialized object (as a _TestCase metaclass method)
    """
    return '{}/{}'.format(cls.group, cls.name)


class _TestCase(type):
    """Metaclass for BaseTest that is used for registering imported tests.

    Attributes:
        cwd (str): path to the directory of the TESTS.py file containing the
            class
        group (str): name of the directory of the TESTS.py file
        testnum (int): test number
        tc_dirname (str): name of the directory created for the test in testdir
    """

    def __init__(cls, name, bases, dct):
        type.__init__(cls, name, bases, dct)

        # globally register class as test case
        # only classes whose names start with 'TEST' are meant to be run
        cls.name = cls.__name__
        if cls.__module__ == '__main__':
            cls.cwd = path.dirname(path.abspath(sys.argv[0]))
        else:
            cls.cwd = cls.__module__

        cls.group = path.basename(cls.cwd)

        if cls.name.startswith('TEST'):
            builtins.testcases.append(cls)
            try:
                cls.testnum = int(cls.name.replace('TEST', ''))
            except ValueError as e:
                print('Invalid test class name {}, should be "TEST[number]"'
                      .format(cls.name))
                raise e

            cls.tc_dirname = cls.group + '_' + str(cls.testnum)

    def __str__(cls):
        return _test_string_repr(cls)


class BaseTest(metaclass=_TestCase):
    """
    Framework base test class. Every test case needs to (directly or
    indirectly) inherit from it. Since this class implements only
    very abstract test behaviour, it is advised for most test cases
    to use Test class inheriting from it.

    Attributes:
        enabled (bool): True if test is meant to be executed, False otherwise.
            Defaults to True.
        force_disabled (bool): True if the test is explicitle excluded from
            execution e.g. in case the test fails permanently.
            Defaults to False.
        ctx (Context): context affiliated with the test
    """
    enabled = True
    force_disabled = False
    labels = []

    def __init__(self):
        self.ctx = None

    def __str__(self):
        return _test_string_repr(self)

    def _execute(self, c):
        """
        Implementation of basic single contextualized test execution workflow.
        Called by the test runner.

        Args:
            c (Context): test context
        """
        self.ctx = c
        start_time = None  # assign in case clean() or setup() fail

        try:
            # pre-execution cleanup
            self.ctx.clean()
            self.clean()

            self.ctx.setup()
            self.setup(c)

            start_time = datetime.now()
            self.run(c)
            self.elapsed = (datetime.now() - start_time).total_seconds()

            self.ctx.check()
            self.check(c)

        except futils.Fail:
            if start_time is not None:
                self.elapsed = (datetime.now() - start_time).total_seconds()
            self._on_fail()
            raise

        except futils.Skip:
            self.ctx.clean()
            self.clean()
            raise

        except sp.TimeoutExpired:
            msg = '{}: {}TIMEOUT{}\t({})'.format(self, futils.Color.RED,
                                                 futils.Color.END,
                                                 self.ctx)
            if start_time is not None:
                self.elapsed = (datetime.now() - start_time).total_seconds()
            raise futils.Fail(msg)

        else:
            self.ctx.clean()
            self.clean()

    def setup(self, ctx):
        """Test setup - not implemented by BaseTest"""
        pass

    def run(self, ctx):
        """
        Main test body, run with specific context provided through
        Context class instance. Needs to be implemented by each test
        """
        raise NotImplementedError('{} does not implement run() method'.format(
            self.__class__))

    def check(self, ctx):
        """Run additional test checks - not implemented by BaseTest"""
        pass

    def clean(self):
        """Test cleanup - not implemented by BaseTest"""
        pass

    def _on_fail(self):
        """Custom behaviour on test fail - not implemented by BaseTest"""
        pass


class Test(BaseTest):
    """
    Generic implementation of BaseTest scaffolding used by particular test
    case classes as a base. In practice, most tests need to inherit from
    this class rather than from Basetest since it implements most commonly
    used test utilities and actions.

    Args:
        test_type (_TestType): test type affiliated with test execution
            length. Defaults to Medium.
        match (bool): True if match files will be checked for test,
            False otherwise. Defaults to True.
        config (Configurator): test execution configuration
        msg (Message): message class instance with logging level
        env (dict): environment variables needed for test execution

    """
    test_type = test_types.Medium
    match = True

    def __init__(self):
        super().__init__()
        self.config = Configurator().config
        self.msg = futils.Message(self.config.unittest_log_level)
        self.log_files = {}

    def _get_utenv(self):
        """Get environment variables values used by C test framework"""
        return {
            'UNITTEST_NAME': str(self),
            'UNITTEST_LOG_LEVEL': str(self.config.unittest_log_level),
            'UNITTEST_NUM': str(self.testnum)
        }

    def _debug_log_env(self):
        """
        Return environment variables that enable logging PMDK debug output
        into log files
        """

        envs = {}
        for libs in LIBS_LIST:
            envs['{}_LOG_LEVEL'.format(libs.upper())] = '3'
            log_file = os.path.join(self.cwd,
                                    '{}_{}.log'.format(libs, self.testnum))
            envs['{}_LOG_FILE'.format(libs.upper())] = log_file
            self.log_files[libs] = log_file
        return envs

    def get_log_files(self):
        """
        Returns names of all log files for given test
        """
        pattern = r'.*[a-zA-Z_]{}\.log'
        log_files = []
        files = os.scandir(self.cwd)
        for file in files:
            match = re.fullmatch(pattern.format(self.testnum), file.name)
            if match:
                log = path.abspath(path.join(self.cwd, file.name))
                log_files.append(log)
        return log_files

    def get_log_file_by_prefix(self, prefix):
        """
        Returns path of a log file with specific prefix and number
        corresponding to self.testnum
        """
        return next(filter(
            lambda x: os.path.basename(x).startswith(
                F"{prefix}_{self.testnum}"), self.get_log_files()
        ))

    def _print_log_files(self):
        """
        Prints all log files for given test
        """
        log_files = self.get_log_files()
        for file in log_files:
            with open(file) as f:
                self.ctx.dump_n_lines(f)

    def _move_log_files(self, ctx):
        """
        Move all log files for given tests
        """
        path = "logs"
        sub_dir = str(ctx).replace(':', '')
        logs_dir = os.path.join(path, sub_dir)
        os.makedirs(logs_dir, exist_ok=True)
        log_files = self.get_log_files()
        for file in log_files:
            shutil.copy2(file, logs_dir)

    def remove_log_files(self):
        """
        Removes log files for given test
        """
        log_files = self.get_log_files()
        for file in log_files:
            os.remove(file)

    def setup(self, ctx):
        """Test setup"""
        self.env = {}
        self.env.update(self._get_utenv())
        self.env.update(self._debug_log_env())
        self.ctx.add_env(self.env)
        # disable shutdown state control for non-pmem (Page granularity)
        if isinstance(ctx.granularity, g.Page):
            pmemobj_conf = os.getenv('PMEMOBJ_CONF', default='')
            if 'sds.at_create=0' not in pmemobj_conf:
                env = dict()
                env['PMEMOBJ_CONF'] = 'sds.at_create=0'
                self.ctx.add_env(env)

        self.remove_log_files()

    def _on_fail(self):
        self._print_log_files()

    def check(self, ctx):
        """Run additional test checks"""
        if self.match:
            self._run_match()
        self._move_log_files(ctx)

    def _run_match(self):
        """Match log files"""
        cwd_listdir = [path.join(self.cwd, f) for f in os.listdir(self.cwd)]

        suffix = '{}.log.match'.format(self.testnum)

        def is_matchfile(f):
            """Match file ends with specific suffix and a char before suffix
            is not a digit"""
            before_suffix = -len(suffix) - 1
            return path.isfile(f) and f.endswith(suffix) and \
                not f[before_suffix].isdigit()

        match_files = filter(is_matchfile, cwd_listdir)
        match_cmd = path.join(ROOTDIR, 'match')

        for mf in match_files:
            cmd = '{} {}'.format(match_cmd, mf)
            proc = sp.run(cmd.split(), stdout=sp.PIPE, cwd=self.cwd,
                          stderr=sp.STDOUT, universal_newlines=True)
            if proc.returncode != 0:
                futils.fail(proc.stdout, exit_code=proc.returncode)
            else:
                self.msg.print_verbose(proc.stdout)



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2024, Intel Corporation
#


import testframework as t
from testframework import granularity as g
import futils as f


@g.require_granularity(g.ANY)
class PMEM2_CONFIG(t.Test):
    test_type = t.Short


class TEST0(PMEM2_CONFIG):
    """test granularity detection with PMEM2_FORCE_GRANULARITY set to page"""
    def run(self, ctx):
        ctx.env['PMEM2_FORCE_GRANULARITY'] = "page"
        ctx.exec(f.get_test_tool_path(ctx.build, "gran_detecto"),
                 '-p', ctx.testdir)


class TEST1(PMEM2_CONFIG):
    """test granularity detection with PMEM2_FORCE_GRANULARITY
    set to cache_line"""
    def run(self, ctx):
        ctx.env['PMEM2_FORCE_GRANULARITY'] = "cache_line"
        ctx.exec(f.get_test_tool_path(ctx.build, "gran_detecto"),
                 '-c', ctx.testdir)


class TEST2(PMEM2_CONFIG):
    """test granularity detection with PMEM2_FORCE_GRANULARITY set to byte"""
    def run(self, ctx):
        ctx.env['PMEM2_FORCE_GRANULARITY'] = "byte"
        ctx.exec(f.get_test_tool_path(ctx.build, "gran_detecto"),
                 '-b', ctx.testdir)


class TEST3(PMEM2_CONFIG):
    """test granularity detection with PMEM2_FORCE_GRANULARITY
    set to CaCHe_Line"""
    def run(self, ctx):
        ctx.env['PMEM2_FORCE_GRANULARITY'] = "CaCHe_Line"
        ctx.exec(f.get_test_tool_path(ctx.build, "gran_detecto"),
                 '-c', ctx.testdir)


class TEST4(PMEM2_CONFIG):
    """test granularity detection with PMEM2_FORCE_GRANULARITY
    set to CACHELINE"""
    def run(self, ctx):
        ctx.env['PMEM2_FORCE_GRANULARITY'] = "CACHELINE"
        ctx.exec(f.get_test_tool_path(ctx.build, "gran_detecto"),
                 '-c', ctx.testdir)



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2020-2023, Intel Corporation
#

import testframework as t


@t.require_ndctl()
class BB_MOCKS_BASIC(t.Test):
    """PART #1 - basic tests"""
    def run(self, ctx):
        test = 'test_basic'
        ctx.exec('pmem2_badblock_mocks', test, ctx.file_type(), ctx.mode())


@t.add_params('file_type', ['reg_file', 'chr_dev'])
@t.add_params('mode', ['no_device'])
class TEST0(BB_MOCKS_BASIC):
    """did not found any matching device"""
    """regular file / character device"""
    pass


@t.add_params('file_type', ['reg_file'])
@t.add_params('mode', ['namespace', 'region'])
class TEST1(BB_MOCKS_BASIC):
    """regular file, namespace mode / region mode"""
    pass


@t.add_params('file_type', ['chr_dev'])
@t.add_params('mode', ['region'])
class TEST2(BB_MOCKS_BASIC):
    """character device, region mode"""
    pass


@t.require_ndctl()
class BB_MOCKS_READ_CLEAR(t.Test):
    """PART #2 - test reading and clearing bad blocks"""
    def run(self, ctx):
        test = 'test_read_clear_bb'
        ctx.exec('pmem2_badblock_mocks',
                 test, ctx.file_type(), ctx.mode(), ctx.bb())


@t.add_params('file_type', ['reg_file'])
@t.add_params('mode', ['namespace', 'region'])
@t.add_params('bb', [1, 2, 3, 4])
class TEST3(BB_MOCKS_READ_CLEAR):
    """regular file, namespace mode / region mode"""
    pass


@t.add_params('file_type', ['chr_dev'])
@t.add_params('mode', ['region'])
@t.add_params('bb', [1, 2, 3, 4])
class TEST4(BB_MOCKS_READ_CLEAR):
    """character device, region mode"""
    pass



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2023, Intel Corporation
#
import futils
import testframework as t
from testframework import granularity as g


@t.require_build(['debug', 'nondebug'])
class EX_LIBPMEM2(t.Test):
    test_type = t.Medium
    file_size = 1 * t.MiB

    offset = str(97 * t.KiB)
    length = str(65 * t.KiB)


class TEST0(EX_LIBPMEM2):

    def run(self, ctx):
        example_path = futils.get_example_path(ctx, 'pmem2', 'basic')
        file_path = ctx.create_non_zero_file(self.file_size, 'testfile0')

        ctx.exec(example_path, file_path)


class TEST1(EX_LIBPMEM2):

    def run(self, ctx):
        example_path = futils.get_example_path(ctx, 'pmem2', 'advanced')
        file_path = ctx.create_non_zero_file(self.file_size, 'testfile0')

        ctx.exec(example_path, file_path, self.offset, self.length)


class TEST2(EX_LIBPMEM2):
    file_size = 16 * t.MiB

    def run(self, ctx):
        example_path = futils.get_example_path(ctx, 'pmem2', 'log')
        file_path = ctx.create_holey_file(self.file_size, 'testfile0')

        args = ['appendv', '4', 'PMDK ', 'is ', 'the best ', 'open source ',
                'append', 'project in the world.', 'dump', 'rewind', 'dump',
                'appendv', '2', 'End of ', 'file.', 'dump']

        ctx.exec(example_path, file_path, *args, stdout_file='out2.log')


class TEST3(EX_LIBPMEM2):

    def run(self, ctx):
        example_path = futils.get_example_path(ctx, 'pmem2', 'redo')
        file_path = ctx.create_holey_file(self.file_size, 'testfile0')
        for x in range(1, 100):
            ctx.exec(example_path, "add", file_path, x, x)
        ctx.exec(example_path, "check", file_path)
        ctx.exec(example_path, "print", file_path, stdout_file='out3.log')


class TEST4(EX_LIBPMEM2):

    def run(self, ctx):
        example_path = futils.get_example_path(ctx, 'pmem2',
                                               'map_multiple_files')

        args = []
        for x in range(1, 10):
            file_path = ctx.create_holey_file(self.file_size,
                                              'testfile{}'.format(x))
            args.append(file_path)

        ctx.exec(example_path, *args, stdout_file='out4.log')


# XXX Disable the test execution under pmemcheck with g.PAGE until the issue
# https://github.com/pmem/pmdk/issues/5641 is fixed.
# additionall test TEST501 has been added to cover non-pmemcheck configs.
class EX_LIBPMEM2_TEST5(EX_LIBPMEM2):

    def run(self, ctx):
        example_path = futils.get_example_path(ctx, 'pmem2', 'unsafe_shutdown')
        file_path = ctx.create_holey_file(self.file_size, 'testfile0')
        ctx.exec(example_path, "write", file_path, "foobar")
        ctx.exec(example_path, "read", file_path, stdout_file='out5.log')


@g.require_granularity(g.CACHELINE, g.BYTE)  # to be removed when fixed
@t.require_valgrind_enabled('pmemcheck')  # to be removed when fixed
class TEST5(EX_LIBPMEM2_TEST5):

    pass


# XXX Disable the test execution with 'memcheck' until the issue:
# https://github.com/pmem/pmdk/issues/5635 is fixed.
# additionall test TEST501 has been added to cover non-pmemcheck configs.
# @t.require_valgrind_disabled('memcheck')  # to be removed when fixed
# @t.require_valgrind_disabled('pmemcheck')
@t.require_valgrind_disabled('pmemcheck', 'memcheck')
class TEST501(EX_LIBPMEM2_TEST5):  # to be removed when fixed

    pass


# XXX disable the test for `memcheck' and 'helgrind'
# until https://github.com/pmem/pmdk/issues/5638 is fixed.
# @t.require_valgrind_disabled('memcheck', 'helgrind')
# XXX disable the test for `drd'
# until https://github.com/pmem/pmdk/issues/5593 is fixed.
# @t.require_valgrind_disabled('drd')
# This test case would require two VALGRIND_SET_CLEAN() calls
# to be added to the "src/examples/libpmem2/ringbuf/ringbuf.c"
# example (see https://github.com/pmem/pmdk/pull/5604)
# in order to pass under pmemcheck, but examples
# do not use valgrind macros on purpose (to avoid unnecessary
# complication), so this test case just should not be run under pmemcheck.
# @t.require_valgrind_disabled('pmemcheck')
# XXX _disabled() can be used only once.
@t.require_valgrind_disabled('memcheck', 'drd', 'pmemcheck', 'helgrind')
class TEST6(EX_LIBPMEM2):

    def run(self, ctx):
        example_path = futils.get_example_path(ctx, 'pmem2', 'ringbuf')
        file_path = ctx.create_holey_file(self.file_size, 'testfile0')
        ctx.exec(example_path, file_path, 10000, 4096, stdout_file='out6.log')



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2020-2022, Intel Corporation
#

from collections import namedtuple

import testframework as t


TC = namedtuple('TC', ['dest', 'src', 'length'])


class Pmem2Memcpy(t.Test):
    test_type = t.Short
    filesize = 4 * t.MiB
    envs0 = ()
    envs1 = ()
    test_cases = (
        # aligned everything
        TC(dest=0, src=0, length=4096),

        # unaligned dest
        TC(dest=7, src=0, length=4096),

        # unaligned dest, unaligned src
        TC(dest=7, src=9, length=4096),

        # aligned dest, unaligned src
        TC(dest=0, src=9, length=4096)
    )

    def run(self, ctx):
        for env in self.envs0:
            ctx.env[env] = '0'
        for env in self.envs1:
            ctx.env[env] = '1'

        if ctx.wc_workaround() == 'on':
            ctx.env['PMEM_WC_WORKAROUND'] = '1'
        elif ctx.wc_workaround() == 'off':
            ctx.env['PMEM_WC_WORKAROUND'] = '0'

        for tc in self.test_cases:
            filepath = ctx.create_holey_file(self.filesize, 'testfile',)
            ctx.exec('pmem2_memcpy', filepath,
                     tc.dest, tc.src, tc.length)


@t.add_params('wc_workaround', ['on', 'off', 'default'])
class TEST0(Pmem2Memcpy):
    pass


@t.add_params('wc_workaround', ['default'])
class TEST1(Pmem2Memcpy):
    envs1 = ("PMEM_NO_MOVNT",)


@t.add_params('wc_workaround', ['default'])
class TEST2(Pmem2Memcpy):
    envs1 = ("PMEM_NO_MOVNT", "PMEM_NO_GENERIC_MEMCPY")


@t.require_architectures('x86_64')
@t.add_params('wc_workaround', ['on', 'off', 'default'])
class TEST3(Pmem2Memcpy):
    envs0 = ("PMEM_AVX512F",)


@t.require_architectures('x86_64')
@t.add_params('wc_workaround', ['on', 'off', 'default'])
class TEST4(Pmem2Memcpy):
    envs0 = ("PMEM_AVX512F", "PMEM_AVX",)


@t.require_architectures('x86_64')
@t.add_params('wc_workaround', ['on', 'off', 'default'])
class TEST5(Pmem2Memcpy):
    envs1 = ("PMEM_MOVDIR64B",)



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019, Intel Corporation
#
"""
Interpreter managing test group specific TESTS.py file execution.
It parses test classes from interpreted file, handles command line arguments
and executes tests using provided configuration.
"""

import importlib.util as importutil
import os
import sys

from testframework import Configurator, get_testcases
from RUNTESTS import TestRunner


def run_testcases():
    """Parse user configuration, run test cases"""
    config = Configurator().config
    testcases = get_testcases()
    runner = TestRunner(config, testcases)
    return runner.run_tests()


def main():
    # Interpreter receives TESTS.py file as first argument
    if len(sys.argv) < 2:
        sys.exit('Provide test file to run')
    testfile = sys.argv[1]

    # Remove TESTS.py file from args, the rest of the args is parsed as a
    # test configuration
    sys.argv.pop(1)

    # import TESTS.py as a module
    testfile_dir = os.path.abspath(os.path.dirname(testfile))
    spec = importutil.spec_from_file_location(testfile_dir, testfile)
    module = importutil.module_from_spec(spec)
    spec.loader.exec_module(module)

    sys.exit(run_testcases())


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2023, Intel Corporation
#

#
# src/test/scope/TESTS.py -- scope tests to check libraries symbols
#

import os
import sys
import subprocess as sp

import futils as ft
import testframework as t
from testframework import granularity as g


def parse_lib(ctx, lib, static=False):
    if sys.platform.startswith('linux'):
        return parse_lib_linux(ctx, lib, static)


def parse_lib_linux(ctx, lib, static):
    if static:
        arg = '-g'
    else:
        arg = '-D'
    cmd = ['nm', arg, lib]
    proc = sp.run(cmd,
                  universal_newlines=True, stdout=sp.PIPE, stderr=sp.STDOUT)
    if proc.returncode != 0:
        raise ft.Fail('command "{}" failed: {}'
                      .format(' '.join(cmd), proc.stdout))

    symbols = []
    for line in proc.stdout.splitlines():
        try:
            # penultimate column of 'nm' output must be either
            # 'T' (defined function) or 'B' (global variable).
            # Example output lines:
            #     000000000003fde4 T pmemobj_create
            #     0000000000000018 B _pobj_cached_pool
            #                      U read
            if line.split()[-2] in 'TB':
                symbols.append(line.split()[-1] + os.linesep)
        except IndexError:
            continue

    symbols.sort()
    return ''.join(symbols)


@t.require_valgrind_disabled('drd', 'helgrind', 'memcheck', 'pmemcheck')
@g.no_testdir()
class Common(t.Test):
    test_type = t.Medium

    checked_lib = ''

    def run(self, ctx):
        static = False
        if str(self.ctx.build) in ['debug', 'nondebug']:
            lib = '{}.so.1'.format(self.checked_lib)
        else:
            static = True
            lib = '{}.a'.format(self.checked_lib)

        libpath = os.path.join(ft.get_lib_dir(ctx), lib)

        log = 'out{}.log'.format(self.testnum)
        out = parse_lib(ctx, libpath, static)
        with open(os.path.join(self.cwd, log), 'w') as f:
            f.write(out)


class TEST2(Common):
    """Check scope of libpmem library (*nix)"""
    labels = ['fault_injection']
    checked_lib = 'libpmem'


class TEST5(Common):
    """Check scope of libpmemobj library (*nix)"""
    labels = ['fault_injection']
    checked_lib = 'libpmemobj'


class TEST6(Common):
    """Check scope of libpmempool library (*nix)"""
    labels = ['fault_injection']
    checked_lib = 'libpmempool'


class TEST13(Common):
    """Check scope of libpmem2 library (*nix)"""
    checked_lib = 'libpmem2'



#!/usr/bin/env python3

# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2023-2024, Intel Corporation

import argparse
import json
import re

from typing import List, Dict, Any

# https://peps.python.org/pep-0589/ requires Python >= 3.8
# from typing import TypedDict
# https://peps.python.org/pep-0613/ requires Python >= 3.10
# from typing import TypeAlias

# Max empirically measured call stacks' size for ndctl_* calls (v78)
NDCTL_CALL_STACK_ESTIMATE = 4608

# Max empirically measured call stacks' size for daxctl_* calls (v78)
DAXCTL_CALL_STACK_ESTIMATE = 7344

PARSER = argparse.ArgumentParser()
PARSER.add_argument('-u', '--stack-usage-file', default='stack_usage.txt')
PARSER.add_argument('-f', '--cflow-output-file', default='cflow.txt')
PARSER.add_argument('-e', '--extra-calls', default='extra_calls.json')
PARSER.add_argument('-a', '--api-file', default='api.txt')
PARSER.add_argument('-w', '--white-list', default='white_list.json')
PARSER.add_argument('-d', '--dump', action='store_true', help='Dump debug files')
PARSER.add_argument('-s', '--dump-all-stacks', action='store_true',
                    help='Dump call_stacks_all.json files. Should be used with caution because the file can be >2GB and dumping takes a significant amount of time.')
PARSER.add_argument('-l', '--filter-lower-limit', type=int, default=0,
        help='Include only call stacks of stack usage above the limit.')
PARSER.add_argument('-i', '--filter-api-file',
        help='Include only call stacks starting from API calls enumerated in the provided file.')
PARSER.add_argument('-t', '--skip-threshold', type=int, default=0,
        help='Ignore non-reachable function if its stack usage <= threshold')

# API: TypeAlias = List[str] # for Python >= 3.10
API = List[str] # for Python < 3.8

# class WhiteList(TypedDict): # for Python >= 3.8
#     not_called: List[str]
WhiteList = Dict[str, List[str]] # for Python < 3.8

# APIFilter: TypeAlias = List[str] # for Python >= 3.10
APIFilter = List[str] # for Python < 3.8

# class StackUsageRecord(TypedDict): # for Python >= 3.8
#     size: int
#     type: str
StackUsageRecord = Dict[str, Any] # for Python < 3.8

# StackUsage: TypeAlias = dict[str, StackUsageRecord] # for Python >= 3.10
StackUsage = Dict[str, StackUsageRecord] # for Python < 3.10

# Calls: TypeAlias = dict[str, list[str]] # for Python >= 3.10
Calls = Dict[str, List[str]] # for Python < 3.10

# RCalls: TypeAlias = Calls # for Python >= 3.10
RCalls = Calls # for Python < 3.10

# class CallStack(TypedDict): # for Python >= 3.8
#     stack: list[str]
#     size: int
CallStack = Dict[str, Any] # for Python < 3.8

DUMP = False

def dump(var, name: str, force: bool = False) -> None:
        global DUMP
        if not force and not DUMP:
                return
        with open(f'{name}.json', 'w') as outfile:
                json.dump(var, outfile, indent = 4)

def load_from_json(file_name: str) -> Any:
        with open(file_name, 'r') as file:
                return json.load(file)

def txt_filter(line: str) -> bool:
        if line == '\n': # drop empty lines
                return False
        if line[0] == '#': # drop comment lines
                return False
        return True

def load_from_txt(file_name: str) -> List[str]:
        with open(file_name, 'r') as file:
                ret = file.readlines()
        ret = list(filter(txt_filter, ret))
        # remove end-line characters
        return [line[:-1] for line in ret]

def parse_stack_usage(stack_usage_file: str) -> StackUsage:
        funcs = {}
        with open(stack_usage_file, 'r') as file:
                for line in file:
                        # 8432 out_common : src/nondebug/libpmem/out.su:out.c dynamic,bounded
                        # 336 __iomem_get_dev_resource : ./build/daxctl/lib/libdaxctl.so.1.2.5.p/.._.._util_iomem.c.su:iomem.c dynamic,bounded
                        # 112 ctl_find_node.isra.5 : src/nondebug/common/ctl.su:ctl.c static
                        found = re.search('([0-9]+) ([a-zA-Z0-9_]+)(.[a-zA-Z0-9.]+)* : ([a-zA-Z0-9.:/_-]+) ([a-z,]+)', line)
                        if found:
                                func = found.group(2)
                                size = int(found.group(1))
                                if func in funcs.keys():
                                        curr_size = funcs[func]['size']
                                        if size != curr_size:
                                                print(f'Warning: Incompatible function records [{func}].size: {size} != {curr_size}. Continue with the biggest reported size.')
                                                if curr_size > size:
                                                        size = curr_size
                                funcs[func] = {'size': size, 'type': found.group(5)}
                        else:
                                print(f'An unexpected line format: {line}')
                                exit(1)
        return funcs

def parse_cflow_output(cflow_output_file: str) -> Calls:
        calls = {}
        call_stack = []
        with open(cflow_output_file, 'r') as file:
                # processing line-by-line
                for line in file:
                        line_copy = line
                        # determine the level of nesting
                        level = 0
                        while line[0] == ' ':
                                level += 1
                                line = line[4:] # remove the prepended spaces
# pmem_memset_persist() <void *pmem_memset_persist (void *pmemdest, int c, size_t len) at pmem.c:731>:
                        found = re.search('^([a-zA-Z0-9_]+)\(\)', line)
                        if not found:
                                print('An unexpected line format:')
                                print(line_copy)
                                exit(1)
                        func = found.group(1)
                        # construct the call stack being currently processed
                        call_stack.insert(level, func)
                        # being level 0 it does not have a caller
                        if level == 0:
                                continue
                        callee = func
                        caller = call_stack[level - 1]
                        if caller in calls.keys():
                                calls[caller].append(callee)
                        else:
                                calls[caller] = [callee]
        # remove duplicate callees
        calls_unique = {}
        for k, v in calls.items():
                v_unique = list(set(v))
                calls_unique[k] = v_unique
        return calls_unique

def dict_extend(dict_, key, values):
        if key not in dict_.keys():
                dict_[key] = values
        else:
                dict_[key].extend(values)
        return dict_

def include_extra_calls(calls: Calls, extra_calls: Calls) -> Calls:
        for k, v in extra_calls.items():
                if k not in calls.keys():
                        calls[k] = v
                else:
                        calls[k].extend(v)
        return calls

def find_api_callers(func: str, calls: Calls, api: API):
        callers = [func]
        visited = [func] # loop breaker
        apis = []
        while len(callers) > 0:
                callers_new = []
                for callee in callers:
                        for k, v in calls.items():
                                # this caller does not call this callee
                                if callee not in v:
                                        continue
                                # it is part of the API
                                if k in visited:
                                        continue
                                if k in api:
                                        apis.append(k)
                                else:
                                        callers_new.append(k)
                                        visited.append(k)
                callers = list(set(callers_new))
                # print(callers)
                # if len(apis) > 0 and len(callers) > 0:
                #         exit(1)
        # if len(apis) == 0:
        #         print(func)
        # assert(len(apis) > 0)
        return apis

def validate(stack_usage: StackUsage, calls:Calls, api: API, white_list: WhiteList, skip_threshold: int) -> None:
        all_callees = [v for _, vs in calls.items()
                       for v in vs]
        all_callees = list(set(all_callees))
        dump(all_callees, 'all_callees')

        # all known functions are expected to be called at least once
        not_called = []
        for k, v in stack_usage.items():
                if k in all_callees:
                        continue
                if k in api:
                        continue
                if k in white_list['not_called']:
                        continue
                if v['size'] <= skip_threshold:
                        continue
                not_called.append(k)
        # Use --dump to see the list of not called functions.
        # Investigate and either fix the call graph or add it to the white list.
        file_name = 'not_called'
        if len(not_called) > 0:
                dump(not_called, file_name, True)
                raise Exception("There are some unreachable functions. See the {}.json file.".format(file_name))
        dump(not_called, file_name)

        # all known functions are expected to be reachable from the API
        no_api_connection = {}
        for k, v in stack_usage.items():
                if k in api:
                        continue
                if k in white_list['not_called']:
                        continue
                if v['size'] <= skip_threshold:
                        continue
                callers = find_api_callers(k, calls, api)
                if len(callers) == 0:
                        no_api_connection[k] = v['size']
        file_name = 'no_api_connection'
        if len(no_api_connection) > 0:
                dump(no_api_connection, file_name, True)
                raise Exception("There are some functions unreachable from API. See the {}.json file.".format(file_name))
        dump(no_api_connection, file_name)

        # dump all zero-sized functions for further assesment
        all_callers = [k for k, _ in calls.items()]
        all_funcs = list(set(all_callees + all_callers))
        zero_funcs = [func for func in all_funcs if func not in stack_usage.keys()]
        zero_funcs = list(filter(lambda f: len(f) > 0, zero_funcs))
        zero_funcs_grep = [func for func in zero_funcs]
        dump(zero_funcs, 'zero_funcs')
        global DUMP
        if DUMP:
                with open('zero_funcs.txt', 'w') as outfile:
                        outfile.write('\n'.join(zero_funcs_grep))

def prepare_rcalls(calls: Calls) -> Calls:
        # preparing a reverse call dictionary
        rcalls: dict[str, list] = {}
        for caller, callees in calls.items():
                for callee in callees:
                        if callee in rcalls.keys():
                                rcalls[callee].append(caller)
                        else:
                                rcalls[callee] = [caller]
        dump(rcalls, 'rcalls')
        return rcalls

def call_stacks_find_the_biggest_starting_at(call_stacks: List[CallStack], api_func: str):
        call_stack_max = {
                'size': 0
        }
        for call_stack in call_stacks:
                if api_func == call_stack['stack'][0]:
                        if call_stack['size'] > call_stack_max['size']:
                                call_stack_max = call_stack
        if 'stack' not in call_stack_max.keys():
                raise Exception(f'No call stack starting at {api_func} has been found.')
        return call_stack_max

def generate_call_stacks_calling_api(call_stack_max: CallStack, callers: List[str], call_stacks: List[CallStack]) -> List[CallStack]:
        call_stacks_new = []
        for call_stack in call_stacks:
                # If the caller is the last position in the stack it means
                # the stack can continue through the API call.
                if call_stack['stack'][-1] in callers:
                        new_call_stack = {
                                'stack': call_stack['stack'] + call_stack_max['stack'],
                                'size': call_stack['size'] + call_stack_max['size']
                        }
                        call_stacks_new.append(new_call_stack)
        return call_stacks_new

def generate_call_stacks_calling_apis(apis_called_internally: List[str], rcalls: RCalls, call_stacks: List[CallStack], debug: bool):
        # list of call stacks which can grow
        call_stacks_called: List[CallStack] = [
                call_stacks_find_the_biggest_starting_at(call_stacks, api)
                        for api in apis_called_internally
        ]
        # list of call stacks which cannot grow any more
        call_stacks_new_end = []
        while len(call_stacks_called) > 0:
                call_stacks_new = []
                for call_stack in call_stacks_called:
                        api_func = call_stack['stack'][0]
                        callers = rcalls[api_func]
                        call_stacks_generated = generate_call_stacks_calling_api(call_stack, callers, call_stacks)
                        call_stacks_new.extend(
                                [cs for cs in call_stacks_generated
                                 if cs['stack'][0] in apis_called_internally])
                        call_stacks_new_end.extend(
                                [cs for cs in call_stacks_generated
                                 if cs['stack'][0] not in apis_called_internally])
                # reduce - keep only the max stack for each API call
                apis = list(set([cs['stack'][0] for cs in call_stacks_new]))
                call_stacks_new_reduced = []
                for api in apis:
                        cs_max = call_stacks_find_the_biggest_starting_at(call_stacks_new, api)
                        call_stacks_new_reduced.append(cs_max)
                if debug:
                        print('end: {}, new: {}, reduced: {}'.format(
                                len(call_stacks_new_end),
                                len(call_stacks_new),
                                len(call_stacks_new_reduced)))
                call_stacks_called = call_stacks_new_reduced
        return call_stacks_new_end

def generate_call_stacks_basic(func: str, stack_usage: StackUsage, rcalls: RCalls, api: API) -> List[CallStack]:
        size = 0
        if func.find("ndctl_", 0) == 0:
                size = NDCTL_CALL_STACK_ESTIMATE
        elif func.find("daxctl_", 0) == 0:
                size = DAXCTL_CALL_STACK_ESTIMATE
        elif func in stack_usage.keys():
                size = stack_usage[func]['size']

        call_stacks = [
                {
                        'stack': [func],
                        'size': size
                }
        ]
        # list of call stacks which cannot grow any more
        call_stacks_completed = []
        # call stack generation loop
        while len(call_stacks) > 0:
                call_stacks_in_progress = []
                for call_stack in call_stacks:
                        callee = call_stack['stack'][0]
                        if callee in api:
                                call_stacks_completed.append(call_stack)
                                continue
                        if callee not in rcalls.keys():
                                call_stacks_completed.append(call_stack)
                                continue
                        for caller in rcalls[callee]:
                                # Note: Breaking the loop does not spoil generating
                                # other call stacks spawning from the same stem.
                                if call_stack['stack'].count(caller) == 2:
                                        continue # loop breaker
                                if caller in stack_usage.keys():
                                        caller_stack_size = int(stack_usage[caller]['size'])
                                else:
                                        caller_stack_size = 0
                                call_stacks_in_progress.append({
                                        'stack': [caller] + call_stack['stack'],
                                        'size': call_stack['size'] + caller_stack_size
                                })
                call_stacks = call_stacks_in_progress
        return call_stacks_completed

def call_stack_key(e):
        return e['size']

def generate_call_stacks(stack_usage: StackUsage, calls: Calls, rcalls: RCalls, api: API, debug: bool = False) -> List[CallStack]:
        # Find all APIs called internally
        apis_called_internally = []
        for api_func in api:
                if api_func in rcalls.keys():
                        apis_called_internally.append(api_func)
        dump(apis_called_internally, 'apis_called_internally')

        internal_api_callers = []
        for api_func in apis_called_internally:
                internal_api_callers.extend(rcalls[api_func])
        internal_api_callers = list(set(internal_api_callers))
        dump(internal_api_callers, 'internal_api_callers')

        if debug:
                print('Basic call stack generation')

        # loop over called functions
        call_stacks = []
        for func in rcalls.keys():
                # If a function calls something else, call stack generation will
                # start from its callees unless it is an internal API caller and
                # does not call non-API functions. For simplicity, it is assumed
                # an internal API caller can start a call stack no matter if it
                # also calls non-API callees.
                if func in calls.keys() and func not in internal_api_callers:
                        continue
                if debug:
                        print(f'Generating call stacks ending at - {func}')
                # Note: generate_call_stacks stop further processing of a call
                # stacks when an API function is found. API functions called
                # internally are handled below.
                call_stacks.extend(generate_call_stacks_basic(func, stack_usage, rcalls, api))
        dump(call_stacks, 'call_stacks_basic')

        if debug:
                print('Call stacks calling APIs generation')

        call_stacks_calling_apis = generate_call_stacks_calling_apis(apis_called_internally, rcalls, call_stacks, debug)
        if debug:
                call_stacks_calling_apis.sort(reverse=True, key=call_stack_key)
        dump(call_stacks_calling_apis, 'call_stacks_calling_apis')

        call_stacks.extend(call_stacks_calling_apis)
        call_stacks.sort(reverse=True, key=call_stack_key)
        return call_stacks

def filter_call_stacks(call_stacks: List[CallStack], api_filter_file: APIFilter) -> List[CallStack]:
        api_filter = load_from_txt(api_filter_file)
        dump(api_filter, 'api_filter')
        print('Load API Filter - done')

        call_stacks_filtered = []
        for call_stack in call_stacks:
                if call_stack['stack'][0] in api_filter:
                        call_stacks_filtered.append(call_stack)
        return call_stacks_filtered

def main():
        args = PARSER.parse_args()
        global DUMP
        DUMP = args.dump # pass the argument value to a global variable

        extra_calls = load_from_json(args.extra_calls)
        print('Load extra calls - done')

        api = load_from_txt(args.api_file)
        dump(api, 'api')
        print('Load API - done')

        white_list = load_from_json(args.white_list)
        print('White List - done')

        stack_usage = parse_stack_usage(args.stack_usage_file)
        # dumping stack_usage.json to allow further processing
        dump(stack_usage, 'stack_usage', True)
        print('Stack usage - done')

        calls = parse_cflow_output(args.cflow_output_file)
        calls = include_extra_calls(calls, extra_calls)
        dump(calls, 'calls')
        print('Function calls - done')

        validate(stack_usage, calls, api, white_list, args.skip_threshold)
        print('Validation - done')

        rcalls = prepare_rcalls(calls)
        print('Reverse calls - done')

        call_stacks = generate_call_stacks(stack_usage, calls, rcalls, api, DUMP)
        dump(call_stacks, 'call_stacks_all', args.dump_all_stacks)
        print('Generate call stacks - done')
        print('Number of call stacks: {}'.format(len(call_stacks)))

        if args.filter_lower_limit > 0:
                def above_limit(call_stack: CallStack) -> bool:
                        return call_stack['size'] > args.filter_lower_limit
                too_big = [call_stack
                           for call_stack in call_stacks
                                if above_limit(call_stack)]
                call_stacks = too_big
                print('Filter out call stacks <= {} - done'.format(args.filter_lower_limit))
                print('Number of call stacks: {}'.format(len(call_stacks)))

        if args.filter_api_file is not None:
                call_stacks = filter_call_stacks(call_stacks, args.filter_api_file)
                print('Filter our call stacks not in the API filter - done')
                print('Number of call stacks: {}'.format(len(call_stacks)))

        dump(call_stacks, 'call_stacks_filtered', True)

if __name__ == '__main__':
        main()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2018-2021, Intel Corporation

import logging

log_levels = ["debug", "info", "warning", "error", "critical"]


class LoggingBase:
    def debug(self, text):
        pass

    def info(self, text):
        pass

    def warning(self, text):
        pass

    def error(self, text):
        pass

    def critical(self, text):
        pass


class DefaultFileLogger(LoggingBase):
    def __init__(self, name="pmreorder", **kwargs):
        logging.basicConfig(**kwargs)
        self.__logger = logging.getLogger(name)

    def debug(self, text):
        self.__logger.debug(text)

    def info(self, text):
        self.__logger.info(text)

    def warning(self, text):
        self.__logger.warning(text)

    def error(self, text):
        self.__logger.error(text)

    def critical(self, text):
        self.__logger.critical(text)


class DefaultPrintLogger(LoggingBase):
    def debug(self, text):
        print("DEBUG:", text)

    def info(self, text):
        print("INFO:", text)

    def warning(self, text):
        print("WARNING:", text)

    def error(self, text):
        print("ERROR:", text)

    def critical(self, text):
        print("CRITICAL:", text)


def get_logger(log_output, log_level=None):
    logger = None
    # check if log_level is valid
    log_level = "warning" if log_level is None else log_level
    numeric_level = getattr(logging, log_level.upper())
    if not isinstance(numeric_level, int):
        raise ValueError("Invalid log level: {}".format(log_level.upper()))

    if log_output is None:
        logger = DefaultPrintLogger()
    else:
        logger = DefaultFileLogger(filename=log_output, level=numeric_level)
    return logger



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2024, Intel Corporation
#


import testframework as t
from testframework import granularity as g


@g.require_granularity(g.ANY)
@t.require_build('nondebug')
@t.require_valgrind_disabled('pmemcheck')
class CORE_LOG(t.BaseTest):
    test_type = t.Short

    def run(self, ctx):
        ctx.exec('core_log_internal', self.test_case)


class TEST0(CORE_LOG):
    test_case = 'test_CORE_LOG'


class TEST1(CORE_LOG):
    test_case = 'test_CORE_LOG_ERROR_LAST'


class TEST2(CORE_LOG):
    test_case = 'test_CORE_LOG_ERROR_W_ERRNO_LAST'


class TEST3(CORE_LOG):
    test_case = 'test_CORE_LOG_W_ERRNO'


class TEST4(CORE_LOG):
    test_case = 'test_CORE_LOG_TRESHOLD'


class TEST5(CORE_LOG):
    test_case = 'test_CORE_LOG_TRESHOLD_DEFAULT'



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2023, Intel Corporation


import testframework as t


@t.require_build(['debug', 'nondebug'])
class BASE(t.Test):
    test_type = t.Medium

    def run(self, ctx):
        filepath = ctx.create_holey_file(16 * t.MiB, 'testfile')
        filepath1 = ctx.create_holey_file(16 * t.MiB, 'testfile1')
        ctx.exec('obj_ulog_size', filepath, filepath1)


@t.require_valgrind_disabled('memcheck', 'pmemcheck')
class TEST0(BASE):
    pass


@t.require_valgrind_enabled('memcheck')
class TEST1(BASE):
    pass


@t.require_valgrind_enabled('pmemcheck')
class TEST2(BASE):
    pass



# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2020-2023, Intel Corporation

"""Test framework constants."""

from os.path import join, abspath, dirname

# List of libraries for logging PMDK debug output
LIBS_LIST = ('pmem', 'pmem2', 'pmemobj', 'pmempool')

# Constant paths to repository elements
ROOTDIR = abspath(join(dirname(__file__), '..'))

DEBUG_LIBDIR = abspath(join(ROOTDIR, '..', 'debug'))
NONDEBUG_LIBDIR = abspath(join(ROOTDIR, '..', 'nondebug'))

HEADER_SIZE = 4096

#
# KiB, MiB, GiB ... -- byte unit prefixes
#
KiB = 2 ** 10
MiB = 2 ** 20
GiB = 2 ** 30
TiB = 2 ** 40
PiB = 2 ** 50


# platform.machine() used for detecting architectures
# gives different values for different systems.
# Key is a normalized value of possible returned architecture values.
NORMALIZED_ARCHS = {
    'x86_64': ('AMD64', 'x86_64'),
    'arm64': ('arm64', 'aarch64'),
    'ppc64el': ('ppc64el', 'ppc64le')
}



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2024, Intel Corporation
#


import testframework as t
from testframework import granularity as g
from os import path


SIGABRT_EXIT_CODE = 134


@g.require_granularity(g.ANY)
# The 'debug' build is chosen arbitrarily to ensure these tests are run only
# once. No dynamic libraries are used nor .static_* builds are available.
@t.require_build('debug')
# (memcheck) When a process forks(3) itself Valgrind can't be stopped from
# following the child process. And since the child process ends with abort(3)
# memory leaks are expected.
# (helgrind, drd) There is no multithreading employed in this test.
# (pmemcheck) It is covered with Bash-based tests.
@t.require_valgrind_disabled('memcheck', 'helgrind', 'drd', 'pmemcheck')
class OBJ_ULOG_ADVANCED(t.Test):
    test_type = t.Short
    test_case = 'test_init_publish_abort_and_verify'
    error_inject = False

    def run(self, ctx):
        testfile = path.join(ctx.testdir, f'testfile{self.testnum}')
        stderr_file = f'err{self.testnum}.log'
        error_inject = 1 if self.error_inject else 0
        # The verify will abort the process when the injected error will be
        # discovered.
        expected_exitcode = SIGABRT_EXIT_CODE if self.error_inject else 0
        ctx.exec('obj_ulog_advanced', self.test_case, testfile, self.slot_num,
                 error_inject, expected_exitcode=expected_exitcode,
                 stderr_file=stderr_file)


class TEST0(OBJ_ULOG_ADVANCED):
    # The number of slots not fully populating a single persistent redo log.
    # Please see the source code for details.
    slot_num = 30


class TEST1(OBJ_ULOG_ADVANCED):
    # The number of slots exactly populating a single persistent redo log.
    # Please see the source code for details.
    slot_num = 40


class TEST2(OBJ_ULOG_ADVANCED):
    # The number of slots between the one used by TEST1 and TEST3.
    slot_num = 50


class TEST3(OBJ_ULOG_ADVANCED):
    # The number of slots exactly populating a persistent shadow log without
    # triggering its growth. Please see the source code for details.
    slot_num = 60


class TEST4(OBJ_ULOG_ADVANCED):
    # The number of slots populating more than a single persistent redo log.
    # It should trigger a persistent shadow log growth.
    # Please see the source code for details.
    slot_num = 70


# Note: Since the injected error value translates to 40 slots and it ought to
# be smaller than the number of populated slots, the error injection only takes
# effect when the number of populated slots is > 40.

class TEST5(TEST2):
    # For details on the injected error please see the source code.
    error_inject = True


class TEST6(TEST3):
    # For details on the injected error please see the source code.
    error_inject = True


class TEST7(TEST4):
    # For details on the injected error please see the source code.
    error_inject = True



# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2023, Intel Corporation

"""Utilities for tests. Meant to be used by test user."""

import platform

import consts as c


def require_architectures(*archs):
    """Enable test only for specified architectures"""
    def wrapped(tc):
        this_arch = platform.machine()

        # normalize this_arch value
        for normalized, possible in c.NORMALIZED_ARCHS.items():
            if this_arch in possible:
                this_arch = normalized

        if this_arch not in archs:
            tc.enabled = False
        return tc

    return wrapped


def DISABLED(**kwargs):
    """
    DISABLE a given test.

    Used as a test class (tc) decorator.
    """

    def wrapped(tc):
        tc.force_disabled = True

    return wrapped



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2020-2023, Intel Corporation
#

import testframework as t


class PMEM2_MAP_PROT(t.Test):
    test_type = t.Short
    filesize = 16 * t.MiB

    def run(self, ctx):
        filepath = ctx.create_holey_file(self.filesize, 'testfile',)
        ctx.exec('pmem2_map_prot', self.test_case, filepath)


class TEST0(PMEM2_MAP_PROT):
    """
    READ/WRITE protection on file opened in read-write mode - should succeed
    """
    test_case = "test_rw_mode_rw_prot"


class TEST1(PMEM2_MAP_PROT):
    """
    READ/WRITE protection on file opened in read-only mode - should fail
    """
    test_case = "test_r_mode_rw_prot"


class TEST4(PMEM2_MAP_PROT):
    """
    READ protection on file opened in read-write mode - should succeed
    """
    test_case = "test_rw_mode_r_prot"


class TEST5(PMEM2_MAP_PROT):
    """
    READ protection on file opened in read-only mode - should succeed
    """
    test_case = "test_r_mode_r_prot"


# PMEM2_PROT_NONE flag is not supported by the CreateFileMapping function.
# This test on purpose performs an "Invalid write"
# which causes Memcheck to fail.
@t.require_valgrind_disabled('memcheck')
class TEST6(PMEM2_MAP_PROT):
    """
    NONE protection on file opened in read-write mode - should succeed
    """
    test_case = "test_rw_mode_none_prot"


@t.require_architectures('x86_64')
# XXX disable the test until https://github.com/pmem/pmdk/issues/5599 is fixed.
@t.require_valgrind_disabled('memcheck')
class TEST7(PMEM2_MAP_PROT):
    """
    READ|EXEC protection on file opened in read|write|exec mode; test runs
    the program, which is put in mapped memory - should succeed
    """
    test_case = "test_rx_mode_rx_prot_do_execute"


class PMEM2_PROT_EXEC(t.Test):
    test_type = t.Short
    filesize = 16 * t.MiB
    is_map_private = 0

    def run(self, ctx):
        filepath = ctx.create_holey_file(self.filesize, 'testfile')
        ctx.exec('pmem2_map_prot', self.test_case,
                 filepath, self.is_map_private)


class TEST8(PMEM2_PROT_EXEC):
    """
    READ|EXEC protection on file opened in read|write|exec mode; test writes
    data to mapped memory - should failed
    """
    test_case = "test_rwx_mode_rx_prot_do_write"


class TEST9(PMEM2_PROT_EXEC):
    """
    READ|EXEC protection on file opened in read|write|exec mode; test writes
    data to mapped memory with MAP_PRIVATE - should failed
    """
    test_case = "test_rwx_mode_rx_prot_do_write"
    is_map_private = 1


@t.require_architectures('x86_64')
class TEST10(PMEM2_PROT_EXEC):
    """
    READ|WRITE|EXEC protection on file opened in read|write|exec mode; test
    runs the program, which is put in mapped memory - should succeed
    """
    test_case = "test_rwx_mode_rwx_prot_do_execute"


@t.require_architectures('x86_64')
class TEST11(PMEM2_PROT_EXEC):
    """
    READ|WRITE protection on file opened in read|write mode; test runs
    the program, which is put in mapped memory with MAP_PRIVATE -
    should succeed
    """
    test_case = "test_rwx_mode_rwx_prot_do_execute"
    is_map_private = 1


@t.require_architectures('x86_64')
class TEST12(PMEM2_PROT_EXEC):
    """
    READ|EXEC protection on file opened in read|write mode; test runs
    the program, which is put in mapped memory - should failed
    """
    test_case = "test_rw_mode_rw_prot_do_execute"


@t.require_architectures('x86_64')
class TEST13(PMEM2_PROT_EXEC):
    """
    READ|EXEC protection on file opened in read|write mode; test runs
    the program, which is put in mapped memory with MAP_PRIVATE - should failed
    """
    test_case = "test_rw_mode_rw_prot_do_execute"
    is_map_private = 1


@t.require_architectures('x86_64')
class TEST14(PMEM2_MAP_PROT):
    """
    READ|EXEC protection on file opened in read|exec|write mode; test
    runs the program, which is put in mapped memory with MAP_PRIVATE -
    should succeed
    """
    test_case = "test_rwx_prot_map_priv_do_execute"


@t.require_devdax(t.DevDax('devdax1'))
class PMEM2_MAP_DEVDAX(t.Test):
    test_type = t.Short

    def run(self, ctx):
        dd = ctx.devdaxes.devdax1
        ctx.exec('pmem2_map_prot', self.test_case, dd.path)


class TEST15(PMEM2_MAP_DEVDAX):
    """
    READ/WRITE protection on device DAX opened in read-write
    mode - should succeed
    """
    test_case = "test_rw_mode_rw_prot"


@t.require_fs_exec
@t.require_architectures('x86_64')
class TEST16(PMEM2_MAP_DEVDAX):
    """
    READ|EXEC protection on device DAX opened in read|write|exec mode; test
    runs the program, which is put in mapped memory - should succeed
    """
    test_case = "test_rx_mode_rx_prot_do_execute"



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2021-2023, Intel Corporation
#

import os
import testframework as t
from testframework import granularity as g
import re


@t.require_admin
@g.require_granularity(g.CACHELINE)
@t.require_ndctl(require_namespace=True)
@g.require_real_pmem()
class PMEM2_BADBLOCK_COUNT(t.Test):
    test_type = t.Short

    def run_test(self, ctx, filepath):
        old_log_level = ctx.env['UNITTEST_LOG_LEVEL']

        # writing to stdout file with UT_OUT requires log level >= 2
        if old_log_level is None or int(old_log_level) < 2:
            ctx.env['UNITTEST_LOG_LEVEL'] = '2'

        bbTool = ctx.badblock
        bbTool.clear_all(filepath)
        bbTool.inject(filepath, 0)

        log_file = 'bb.log'
        ctx.exec('pmem2_badblock', self.test_case, filepath,
                 stdout_file=log_file)

        bb_log_path = os.path.join(self.cwd, log_file)

        log_content = open(bb_log_path).read()
        bb_line = re.findall("BB: .*", log_content, re.MULTILINE).pop()

        # get only the log content after BB log header
        bb_log_header = "BB: "
        bb = int(bb_line[len(bb_log_header):])

        bb_from_tool = bbTool.get_count(filepath)
        if bb_from_tool != bb:
            t.futils.fail("BB count mismatch (tool: {}, pmem2: {})"
                          .format(bb_from_tool, bb))
        bbTool.clear_all(filepath)

        # restore previous log level
        ctx.env['UNITTEST_LOG_LEVEL'] = old_log_level


@t.require_admin
@g.require_granularity(g.CACHELINE)
@t.require_ndctl(require_namespace=True)
@g.require_real_pmem()
class PMEM2_BADBLOCK(t.Test):
    test_type = t.Short

    def run_test(self, ctx, filepath):
        bbTool = ctx.badblock
        bbTool.clear_all(filepath)
        bbTool.inject(filepath, 0)

        ctx.exec('pmem2_badblock', self.test_case, filepath)

        bbTool.clear_all(filepath)


# XXX - https://github.com/pmem/pmdk/issues/5636
@t.DISABLED()
class TEST0(PMEM2_BADBLOCK_COUNT):
    """
    compares the number of bad blocks returned by pmem2 and ndctl on fsdax
    """
    test_case = "test_pmem2_badblock_count"

    def run(self, ctx):
        filepath = ctx.create_holey_file(4 * t.KiB, 'testfile')
        self.run_test(ctx, filepath)


# XXX - https://github.com/pmem/pmdk/issues/5636
@t.DISABLED()
@t.require_devdax(t.DevDax('devdax1'))
class TEST1(PMEM2_BADBLOCK_COUNT):
    """
    compares the number of bad blocks returned by pmem2 and ndctl on devdax
    """
    test_case = "test_pmem2_badblock_count"

    def run(self, ctx):
        ddpath = ctx.devdaxes.devdax1.path
        self.run_test(ctx, ddpath)


# XXX - https://github.com/pmem/pmdk/issues/5636
@t.DISABLED()
class TEST2(PMEM2_BADBLOCK):
    """test mcsafe read operation with encountered badblock"""
    test_case = "test_pmem2_src_mcsafe_badblock_read"

    def run(self, ctx):
        filepath = ctx.create_holey_file(4 * t.KiB, 'testfile')
        self.run_test(ctx, filepath)


# XXX - https://github.com/pmem/pmdk/issues/5636
@t.DISABLED()
class TEST3(PMEM2_BADBLOCK):
    """test mcsafe write operation with encountered badblock"""
    test_case = "test_pmem2_src_mcsafe_badblock_write"

    def run(self, ctx):
        filepath = ctx.create_holey_file(4 * t.KiB, 'testfile')
        self.run_test(ctx, filepath)


# XXX - https://github.com/pmem/pmdk/issues/5636
@t.DISABLED()
@t.require_devdax(t.DevDax('devdax1'))
class TEST4(PMEM2_BADBLOCK):
    """test mcsafe read operation with encountered badblock on devdax"""
    test_case = "test_pmem2_src_mcsafe_badblock_read"

    def run(self, ctx):
        ddpath = ctx.devdaxes.devdax1.path
        self.run_test(ctx, ddpath)


# XXX - https://github.com/pmem/pmdk/issues/5636
@t.DISABLED()
@t.require_devdax(t.DevDax('devdax1'))
class TEST5(PMEM2_BADBLOCK):
    """test mcsafe write operation with encountered badblock on devdax"""
    test_case = "test_pmem2_src_mcsafe_badblock_write"

    def run(self, ctx):
        ddpath = ctx.devdaxes.devdax1.path
        self.run_test(ctx, ddpath)



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2020-2023, Intel Corporation
#


import testframework as t


@t.require_ndctl()
class TEST0(t.Test):
    test_type = t.Short

    def run(self, ctx):
        testfile1 = ctx.create_holey_file(2 * t.MiB, 'testfile1')
        ctx.exec('pmem2_source_numa', 'test_get_numa_node',
                 testfile1, 0)


@t.require_ndctl()
@t.require_devdax(t.DevDax('devdax', alignment=2 * t.MiB))
class TEST1(t.Test):
    test_type = t.Short

    def run(self, ctx):
        dd = ctx.devdaxes.devdax
        ctx.exec('pmem2_source_numa', 'test_get_numa_node',
                 dd.path, 0)



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2024, Intel Corporation
#


import testframework as t
from testframework import granularity as g


@g.require_granularity(g.ANY)
@t.require_build('nondebug')
@t.require_valgrind_disabled('pmemcheck', 'memcheck')
class CORE_LOG_MT(t.BaseTest):
    test_type = t.Short

    def run(self, ctx):
        ctx.exec('core_log_threshold_mt', self.test_case)


class TEST0(CORE_LOG_MT):
    test_case = 'test_threshold_set_get'


@t.require_valgrind_enabled('helgrind')
class TEST1(TEST0):
    pass


@t.require_valgrind_enabled('drd')
class TEST2(TEST0):
    pass


class TEST3(CORE_LOG_MT):
    test_case = 'test_threshold_aux_set_get'


@t.require_valgrind_enabled('helgrind')
class TEST4(TEST3):
    pass


@t.require_valgrind_enabled('drd')
class TEST5(TEST3):
    pass



# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2020, Intel Corporation
#

"""Test framework public interface.

All test framework modules that are meant to be used by the test developer
are collected here. The TESTS.py file can conveniently import it:

import testframework as t
"""


# add unittest modules to the visible modules path
import sys
from os import path

# Don't create cache files on the import of test modules.
# Improvement using bytecode in python tf code is insignificant.
# It is workaround for a pycache clobber issue in our tangled Makefiles.
sys.dont_write_bytecode = True
sys.path.insert(1, path.abspath(path.join(path.dirname(__file__), 'unittest')))

# flake8 issues silenced:
# E402 - import statements not at the top of the file because of adding
# directory to path
# F401, F403 - testframework.py does not use imported names, only passes them
# down and in most cases needs to pass down all of them - hence import with '*'

from basetest import BaseTest, Test, get_testcases  # noqa: E402, F401
from context import *  # noqa: E402, F401, F403
from configurator import *  # noqa: E402, F401, F403
from valgrind import *  # noqa: E402, F401, F403
from utils import *  # noqa: E402, F401, F403
from poolset import *  # noqa: E402, F401, F403
from builds import *  # noqa: E402, F401, F403
from devdax import *  # noqa: E402, F401, F403
from test_types import *  # noqa: E402, F401, F403
from requirements import *  # noqa: E402, F401, F403
from unsafe_shutdown import *  # noqa: E402, F401, F403
import granularity  # noqa: E402, F401



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2024, Intel Corporation
#


import testframework as t


class TEST0(t.Test):
    test_type = t.Short

    def run(self, ctx):
        testfile = t.path.join(ctx.testdir, 'testfile')
        ctx.exec('obj_tx_user_data', testfile)



#!usr/bin/env python3
#
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2018-2023, Intel Corporation

"""
This module includes tests that check if all packages in PMDK library are
built correctly.
Tests check:
-if all required packages are built,
-if built packages are consistent with names of libraries read from
 .so files and other elements (tools and "PMDK").

Required arguments:
-r <PMDK_path>    the PMDK library root path.

Optional arguments:
--skip-daxio      to be set if daxio was not built (e.g., if NDCTL_ENABLE was set to 'n')
"""

from os import listdir, path, linesep
from collections import namedtuple
import distro
import unittest
import xmlrunner
import sys
import re

PACKAGES_INFO = namedtuple('packages_info',
                           'basic devel debug debuginfo debug_debuginfo')
PMDK_VERSION = ''
SYSTEM_ARCHITECTURE = ''


def get_package_version_and_system_architecture(pmdk_path):
    """
    Returns packages version and system architecture from names of directories
    from packages directory.
    """
    os_distro=distro.id()
    if os_distro != 'ubuntu':
        pkg_directory = path.join(pmdk_path, 'rpm')
    else:
        pkg_directory = path.join(pmdk_path, 'dpkg')

    version = ''
    architecture = ''
    for elem in listdir(pkg_directory):
        if os_distro != 'ubuntu':
            if '.src.rpm' in elem:
                # looks for the version number of rpm package in rpm package name
                version = re.search(r'[\s]*pmdk-([\S]+).src.rpm', elem).group(1)
            else:
                architecture = elem

        else:
            if '.changes' in elem:
                # looks for the version number of packages in package name
                version = re.search(r'pmdk_*(.+)_(.+).changes', elem).group(1)
                architecture = re.search(r'pmdk_*(.+)_(.+).changes', elem).group(2)
    return version, architecture


def get_built_packages(pmdk_pkg_path):
    """
    Returns built packages.
    """
    packages = listdir(pmdk_pkg_path)
    return packages


def get_libraries_names_from_so_files(pmdk_path, is_pmdk_debuginfo):
    """
    Returns names of libraries from .so files, and information which packages
    should be built for individual libraries.
    """
    libraries_from_so_files = dict()
    path_to_so_files = path.join(pmdk_path, 'src', 'nondebug')

    for elem in listdir(path_to_so_files):
        if elem.endswith('.so') and elem.startswith('lib'):
            library_name = elem.split('.')[0]
            if is_pmdk_debuginfo:
                libraries_from_so_files[library_name] =\
                    PACKAGES_INFO(basic=True, devel=True, debug=True,
                                  debuginfo=False, debug_debuginfo=False)
            else:
                libraries_from_so_files[library_name] =\
                    PACKAGES_INFO(basic=True, devel=True, debug=True,
                                  debuginfo=True, debug_debuginfo=True)
    return libraries_from_so_files


def get_names_of_packages(packages_info):
    """
    Returns names of packages, that should be built.
    """
    packages = []
    pkg_ext = ''
    separator = ''
    os_distro=distro.id()
    if os_distro != 'ubuntu':
        types = ['-', '-debug-', '-devel-', '-debuginfo-', '-debug-debuginfo-']
        pkg_ext = '.rpm'
        separator ='.'

    elif os_distro == 'ubuntu':
        types = ['_', '-dev_']
        pkg_ext = '.deb'
        separator = '_'

    for elem in packages_info:
        sets_of_information = zip(packages_info[elem], types)
        for kit in sets_of_information:
            if kit[0] and os_distro == 'opensuse':
                if elem.startswith('lib') and (kit[1] == '-' or kit[1] == '-debuginfo-'):
                    package_name = elem + str(1) + kit[1] + PMDK_VERSION + separator + \
                        SYSTEM_ARCHITECTURE + pkg_ext
                    packages.append(package_name)
                elif kit[0]:
                    package_name = elem + kit[1] + PMDK_VERSION + separator + \
                        SYSTEM_ARCHITECTURE + pkg_ext
                    packages.append(package_name)

            elif kit[0] and not os_distro == 'opensuse':
                package_name = elem + kit[1] + PMDK_VERSION + separator +\
                    SYSTEM_ARCHITECTURE + pkg_ext
                packages.append(package_name)
    return packages


def check_existence_of_pmdk_debuginfo_package(pmdk_debuginfo_package_name, built_packages):
    """
    Checks if 'pmdk-debuginfo' package is built
    """
    is_pmdk_debuginfo_package = False
    if pmdk_debuginfo_package_name in built_packages:
        is_pmdk_debuginfo_package = True
    return is_pmdk_debuginfo_package


def find_missing_packages(packages_path, pmdk_path, pmdk_debuginfo_package_name):
    """
    Checks if names of built packages are the same as names of packages,
    which should be built and returns missing packages. Tools are taken
    into account.
    """
    built_packages = get_built_packages(packages_path)
    is_pmdk_debuginfo =\
        check_existence_of_pmdk_debuginfo_package(pmdk_debuginfo_package_name, built_packages)
    tools = {
        'pmempool': PACKAGES_INFO(basic=True, devel=False, debug=False,
                                  debuginfo=True, debug_debuginfo=False),
        'pmreorder': PACKAGES_INFO(basic=True, devel=False, debug=False,
                                   debuginfo=False, debug_debuginfo=False)
    }
    if not skip_daxio:
        tools['daxio'] = PACKAGES_INFO(basic=True, devel=False, debug=False,
                                       debuginfo=True, debug_debuginfo=False)

    tools_packages = get_names_of_packages(tools)
    missing_tools_packages = [
        elem for elem in tools_packages if elem not in built_packages]
    libraries = get_libraries_names_from_so_files(pmdk_path, is_pmdk_debuginfo)
    library_packages = get_names_of_packages(libraries)
    missing_library_packages = [
        elem for elem in library_packages if elem not in built_packages]
    missing_packages = missing_library_packages + missing_tools_packages
    return missing_packages


def find_missing_libraries_and_tools(packages_path, pmdk_path, pmdk_debuginfo_package_name, library_name_pattern):
    """
    Checks if names of functions from .so files are the same as names of
    functions extracted from built packages and returns missing functions.
    Others pkg (tools and "PMDK") are taken into account.
    """
    others_pkg = ['pmempool', 'daxio', 'pmdk', 'pmreorder']
    built_packages = get_built_packages(packages_path)
    is_pmdk_debuginfo =\
        check_existence_of_pmdk_debuginfo_package(pmdk_debuginfo_package_name, built_packages)
    libraries = get_libraries_names_from_so_files(pmdk_path, is_pmdk_debuginfo)

    missing_elements = []
    # looks for the name of library/others in package name
    for elem in listdir(packages_path):
        library_name = re.search(library_name_pattern, elem).group(1)
        if library_name.endswith('1'):
            library_name = library_name.replace('1','')
        if library_name not in libraries.keys() and library_name not in\
                others_pkg and library_name not in missing_elements:
            missing_elements.append(library_name)
    return missing_elements


def parse_argument(argument_option):
    """
    Parses an option from the command line.
    """
    index = sys.argv.index(argument_option)
    try:
        argument = sys.argv[index+1]
    except IndexError:
        print('ERROR: Invalid argument!')
        print(__doc__)
        print(unittest.main.__doc__)
    else:
        sys.argv.pop(index)
        sys.argv.pop(index)
        return argument


class TestBuildPackages(unittest.TestCase):

    def test_completeness_of_built_packages(self):
        """
        Checks if all packages are built.
        """
        missing_packages =\
            find_missing_packages(packages_path, pmdk_path, pmdk_debuginfo_package_name)
        error_msg = linesep + 'List of missing packages:'

        for package in missing_packages:
            error_msg += linesep + package
        self.assertFalse(missing_packages, error_msg)

    def test_completeness_of_name_of_libraries_and_tools(self):
        """
        Checks if names of functions from .so files and other elements (tools
        and "PMDK") are the same as functions/other elements extracted from
        the name of built packages.
        """
        os_distro=distro.id()
        missing_elements =\
        find_missing_libraries_and_tools(packages_path, pmdk_path, pmdk_debuginfo_package_name,library_name_pattern)
        error_msg = linesep +\
            'List of missing libraries and other elements (tools and "PMDK"):'
        for elem in missing_elements:
            error_msg += linesep + elem

        self.assertFalse(missing_elements, error_msg)


if __name__ == '__main__':
    path_argument = '-r'
    daxio_build_argument = "--skip-daxio"
    skip_daxio = False
    if '-h' in sys.argv or '--help' in sys.argv:
        print(__doc__)
        unittest.main()
    elif path_argument in sys.argv:
        pmdk_path = parse_argument(path_argument)
        if daxio_build_argument in sys.argv:
            skip_daxio = True
            index = sys.argv.index(daxio_build_argument)
            sys.argv.pop(index)
        if pmdk_path:
            PMDK_VERSION, SYSTEM_ARCHITECTURE =\
                get_package_version_and_system_architecture(pmdk_path)
            os_distro=distro.id()
            if os_distro != 'ubuntu':
                packages_path = path.join(pmdk_path, 'rpm', SYSTEM_ARCHITECTURE)
                pmdk_debuginfo_package_name = 'pmdk-debuginfo-' + PMDK_VERSION + '.' + SYSTEM_ARCHITECTURE + '.rpm'
                library_name_pattern = r'[\s]*([2a-zA-Z+\d]+)-'
            else:
                packages_path = path.join(pmdk_path, 'dpkg')
                pmdk_debuginfo_package_name = 'pmdk-debuginfo-' + PMDK_VERSION + '.' + '.deb'
                library_name_pattern = r'^([2a-zA-Z]+)[-_].*$'
            unittest.main(
                testRunner=xmlrunner.XMLTestRunner(output='test-reports'),
                # these make sure that some options that are not applicable
                # remain hidden from the help menu.
                failfast=False, buffer=False, catchbreak=False)
    else:
        print(__doc__)
        print(unittest.main.__doc__)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2023, Intel Corporation
#

"""Device dax context classes and utilities"""

import copy
import itertools
import os

import context as ctx
import futils
import requirements as req
import tools


class DevDax():
    """
    Class representing dax device and its parameters.

    Attributes:
        name (str): name of the DevDax object by which it is
            accessible from DaxDevices class
        path (str): path to device
        _req_alignment (int): required dax device alignment
        _req_min_size (int): required dax device minimal size
        _req_max_size (int): required dax device maximal size
        assigned (bool): True, if object was successfully assigned
            to real dax device that fulfills its requirements,
            False otherwise

    """
    def __init__(self, name, alignment=None, min_size=None, max_size=None):
        self.name = name
        self.path = None
        self._req_alignment = alignment
        self._req_min_size = min_size
        self._req_max_size = max_size
        self.assigned = False

    def __str__(self):
        return self.name

    def try_assign(self, path):
        """
        Try assigning to real dax device, identified by its path,
        provided it meets defined requirements. In case of success, set DevDax
        object attributes (like size and/or alignment) to the real dax device
        values and return True. Return False otherwise.
        """
        ndctl = tools.Ndctl()

        p_size = ndctl.get_dev_size(path)
        p_align = ndctl.get_dev_alignment(path)

        if self._req_min_size and p_size < self._req_min_size:
            return False
        if self._req_max_size and p_size > self._req_max_size:
            return False
        if self._req_alignment and p_align != self._req_alignment:
            return False

        self.path = path
        self.size = p_size
        self.alignment = p_align
        self.assigned = True
        return True


class DevDaxes():
    """
    Dax device context element class representing a set of dax devices required
    for the test.

    Attributes:
        dax_devices (tuple): set of dax devices (DevDax objects)
            which are required by the test.
            Each DevDax is accessible from DaxDevices
            as an attribute through its name

    """
    check_fs_exec = False

    def __init__(self, *dax_devices):
        self.dax_devices = tuple(dax_devices)
        for dd in dax_devices:
            setattr(self, dd.name, dd)

    def setup(self, **kwargs):
        # DevDax tests always require ndctl
        req.Requirements().check_ndctl_enable()
        req.Requirements().check_ndctl()

        tools = kwargs['tools']
        for dd in self.dax_devices:
            proc = tools.pmemdetect('-d', dd.path)
            if proc.returncode != 0:
                raise futils.Fail('checking {} with pmemdetect failed:{}{}'
                                  .format(dd.path, os.linesep, proc.stdout))

            if self.check_fs_exec:
                exec_allowed = tools.mapexec(dd.path)
                if exec_allowed.returncode != 1:
                    raise futils.Skip('dax device {} has no exec rights for'
                                      ' mmap'.format(dd.path))

    def __str__(self):
        return 'devdax'

    @classmethod
    def filter(cls, config, msg, tc):
        """
        Initialize DevDaxes class for the test to be run
        based on configuration and test requirements.

        Args:
            config: configuration as returned by Configurator class
            msg (Message): level based logger class instance
            tc (BaseTest): test case, from which the dax device run
                requirements are obtained

        Returns:
            Initialized DevDaxes class as single element list for type
            compliance

        """
        dax_devices, _ = ctx.get_requirement(tc, 'devdax', ())
        cls.check_fs_exec, _ = ctx.get_requirement(tc, 'require_fs_exec',
                                                   None)
        req_real_pmem, _ = ctx.get_requirement(tc, 'require_real_pmem', False)

        if not dax_devices:
            return ctx.NO_CONTEXT

        if not config.device_dax_path:
            raise futils.Skip('No dax devices defined in testconfig')

        if len(dax_devices) > len(config.device_dax_path):
            raise futils.Skip('Not enough dax devices defined in testconfig '
                              '({} needed)'.format(len(dax_devices)))

        ndctl = tools.Ndctl()
        for c in config.device_dax_path:
            if not ndctl.is_devdax(c):
                raise futils.Fail('{} is not a dax device'.format(c))

        assigned = _try_assign_by_requirements(config.device_dax_path,
                                               dax_devices)

        # filter out the emulated devdaxes from the assigned ones
        if req_real_pmem:
            assigned_filtered = []

            for dd in assigned:
                if not ndctl.is_emulated(dd.path):
                    assigned_filtered.append(dd)

            assigned = tuple(assigned_filtered)

        if not assigned:
            raise futils.Skip('Dax devices in test configuration do not '
                              'meet test requirements')

        return [DevDaxes(*assigned)]


def _try_assign_by_requirements(configured, required):
    """
    Try assigning dax devices defined as paths in test configuration to
    dax device objects taking their requirements into account.
    Return a sequence of all requirement objects if they were
    successfully assigned to existing dax, otherwise return None.
    Since the order in which requirement objects are tried to be
    assigned may affect the final outcome, all permutations are checked.
    """
    permutations = itertools.permutations(required)
    for p in permutations:
        conf_copy = configured[:]
        req_copy = copy.deepcopy(p)
        for dd in req_copy:
            for i, c in enumerate(conf_copy):
                if dd.try_assign(c):
                    conf_copy.pop(i)
                    break

            if not dd.assigned:
                # at least one device dax requirement cannot be assigned
                # to any of devices defined in the configuration,
                # try another permutation
                break

        if all(dd.assigned for dd in req_copy):
            return req_copy
    return None


def require_devdax(*dax_devices, **kwargs):
    """
    Add requirement to run test on specified dax devices.

    Used as a test class decorator.

    Args:
        dax_devices: variable length list of initialized DevDax objects
        kwargs: optional keyword arguments
    """
    def wrapped(tc):
        ctx.add_requirement(tc, 'devdax', tuple(dax_devices), **kwargs)
        return tc
    return wrapped


def require_fs_exec(tc):
    """
    Disable test if exec access for device is not permitted
    """
    ctx.add_requirement(tc, 'require_fs_exec', True)
    return tc



# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2023, Intel Corporation
#
"""Build context classes"""

import context as ctx
import futils
import consts as c


class Build(metaclass=ctx.CtxType):
    """Base and factory class for standard build classes"""
    exesuffix = ''

    @classmethod
    def filter(cls, config, msg, tc):
        req_builds, kwargs = ctx.get_requirement(tc, 'build', None)

        builds = []
        for b in ctx.filter_contexts(config.build, req_builds):
            try:
                builds.append(b(**kwargs))
            except futils.Skip as s:
                msg.print('{}: SKIP: {}'.format(tc, s))

        return builds


class Debug(Build):
    """Set this context for a debug build"""
    is_preferred = True

    def __init__(self):
        self.libdir = c.DEBUG_LIBDIR
        self.env = {'LD_LIBRARY_PATH': self.libdir}


class NonDebug(Build):
    """Set this context for a nondebug build"""
    is_preferred = True

    def __init__(self):
        self.libdir = c.NONDEBUG_LIBDIR
        self.env = {'LD_LIBRARY_PATH': self.libdir}


class Static_Debug(Build):
    """Set this context for a static_debug build"""

    def __init__(self):
        self.exesuffix = '.static_debug'
        self.libdir = c.DEBUG_LIBDIR


class Static_Nondebug(Build):
    """Set this context for a static_nondebug build"""

    def __init__(self):
        self.exesuffix = '.static_nondebug'
        self.libdir = c.NONDEBUG_LIBDIR


def require_build(build, **kwargs):
    def wrapped(tc):
        builds = ctx.str_to_ctx_common(build, Build)
        ctx.add_requirement(tc, 'build', builds, **kwargs)
        return tc
    return wrapped



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2024, Intel Corporation
#


import testframework as t
from testframework import granularity as g


@g.require_granularity(g.ANY)
@t.require_valgrind_disabled('pmemcheck')
# The 'nondebug' build is chosen arbitrarily to ensure these tests are run only
# once. No dynamic libraries are used nor .static_* builds are available.
@t.require_build('nondebug')
class CORE_LOG(t.BaseTest):
    test_type = t.Short

    def run(self, ctx):
        ctx.exec('core_log_no_func', self.test_case)


class TEST0(CORE_LOG):
    test_case = 'test_no_log_function'


class TEST1(CORE_LOG):
    test_case = 'test_no_log_function_CORE_LOG_LEVEL_ERROR_LAST'



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2024, Intel Corporation


import testframework as t


class BASIC(t.Test):
    test_type = t.Medium

    def run(self, ctx):
        filepath = ctx.create_holey_file(16 * t.MiB, 'testfile1')
        ctx.exec('obj_basic_integration', filepath)


@t.require_valgrind_disabled('memcheck')
class TEST14(BASIC):
    pass


@t.require_valgrind_enabled('pmemcheck')
class TEST15(BASIC):
    pass



# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2023, Intel Corporation
#
"""Valgrind handling tools"""

import sys
import re
import subprocess as sp
from enum import Enum, unique
from os import path

import context as ctx
import futils
from consts import ROOTDIR


DISABLE = -1
ENABLE = 1
AUTO = 0

_IGNORED = (
    "WARNING: Serious error when reading debug info",
    "When reading debug info from ",
    "Ignoring non-Dwarf2/3/4 block in .debug_info",
    "Last block truncated in .debug_info; ignoring",
    "parse_CU_Header: is neither DWARF2 nor DWARF3 nor DWARF4",
    "brk segment overflow",
    "see section Limitations in user manual",
    "Warning: set address range perms: large range",
    "further instances of this message will not be shown",
    "get_Form_contents: DW_FORM_GNU_strp_alt used, but no alternate .debug_str"
)


@unique
class _Tool(Enum):
    """
    Enumeration of valid Valgrind tool options.

    NONE is a virtual tool option denoting that valgrind
    is not used for test execution.
    """
    MEMCHECK = 1
    PMEMCHECK = 2
    HELGRIND = 3
    DRD = 4
    NONE = 5

    def __str__(self):
        return self.name.lower()

    def __bool__(self):
        return self != self.NONE


TOOLS = tuple(t for t in _Tool if t != _Tool.NONE)

MEMCHECK = _Tool.MEMCHECK
PMEMCHECK = _Tool.PMEMCHECK
HELGRIND = _Tool.HELGRIND
DRD = _Tool.DRD
NONE = _Tool.NONE


class Valgrind:
    """Valgrind management context element class

    Attributes:
        tool (_Tool): selected valgrind tool
        tool_name (str): tool name as string
        cwd (str): path to the test cwd
            (i. e. the directory in which test .py file resides)
        log_file (str): path to valgrind output log file
        valgrind_exe (str): path to valgrind command
        opts (list): list of valgrind command line options

    """

    def __init__(self, tool, cwd, testnum, force_enable):
        self.tool = NONE if tool is None or tool == 'none' else tool
        self.tool_name = self.tool.name.lower()
        self.cwd = cwd

        log_file_name = '{}{}.log'.format(self.tool.name.lower(), testnum)
        self.log_file = path.join(cwd, log_file_name)

        if self.tool == NONE:
            self.valgrind_exe = None
        else:
            self.valgrind_exe = self._get_valgrind_exe(force_enable)

        if self.valgrind_exe is None:
            return

        self.verify()

        self.opts = []

        self.add_suppression('ld.supp')

        if tool == MEMCHECK:
            self.add_suppression('memcheck-libunwind.supp')
            self.add_suppression('memcheck-ndctl.supp')
            self.add_suppression('memcheck-dlopen.supp')

        # Before Skylake, Intel CPUs did not have clflushopt instruction, so
        # pmem_flush and pmem_persist both translated to clflush.
        # This means that missing pmem_drain after pmem_flush could only be
        # detected on Skylake+ CPUs.
        # This option tells pmemcheck to expect fence (sfence or
        # VALGRIND_PMC_DO_FENCE client request, used by pmem_drain) after
        # clflush and makes pmemcheck output the same on pre-Skylake and
        # post-Skylake CPUs.
        elif tool == PMEMCHECK:
            self.add_opt('--expect-fence-after-clflush=yes')

        elif tool == HELGRIND:
            self.add_suppression('helgrind-log.supp')

        elif tool == DRD:
            self.add_suppression('drd-log.supp')

    def __str__(self):
        return self.tool.name.lower()

    def __bool__(self):
        return self.tool != NONE

    @classmethod
    def filter(cls, config, msg, tc):
        """
        Acquire Valgrind tool for the test to be run with.
        Takes into account configuration 'force-enable' options,
        and Valgrind tools enabled or disabled by test requirements.

        Args:
            config: configuration as returned by Configurator class
            msg (Message): level based logger class instance
            tc (BaseTest): test case, from which the Valgrind
                requirements are obtained

        Returns:
            list of initialized Valgrind tool classes with which the test
            should be run

        """
        vg_tool, kwargs = ctx.get_requirement(tc, 'enabled_valgrind', NONE)
        disabled, _ = ctx.get_requirement(tc, 'disabled_valgrind', ())

        if config.force_enable:
            if vg_tool and vg_tool != config.force_enable:
                raise futils.Skip(
                    "test enables the '{}' Valgrind tool while "
                    "execution configuration forces '{}'"
                    .format(vg_tool, config.force_enable))

            elif config.force_enable in disabled:
                raise futils.Skip(
                      "forced Valgrind tool '{}' is disabled by test"
                      .format(config.force_enable))

            else:
                vg_tool = config.force_enable
        return [cls(vg_tool, tc.cwd, tc.testnum,
                    config.force_enable is not None, **kwargs), ]

    @property
    def cmd(self):
        """
        Return Valgrind command with specified arguments
        as a subprocess compliant list.
        """
        if self.tool == NONE:
            return []

        cmd = [self.valgrind_exe, '--tool={}'.format(self.tool_name),
               '--log-file={}'.format(self.log_file)] + self.opts
        return cmd

    def setup(self, memcheck_check_leaks=True, **kwargs):
        if self.tool == MEMCHECK and memcheck_check_leaks:
            self.add_opt('--leak-check=full')

    def check(self, **kwargs):
        self.validate_log()

    def _get_valgrind_exe(self, force_enable):
        """
        On some systems "valgrind" is a shell script that calls the actual
        executable "valgrind.bin".
        The wrapper script does not work well with LD_PRELOAD so we want
        to call Valgrind directly.
        """
        try:
            out = sp.check_output('which valgrind', shell=True,
                                  universal_newlines=True, stderr=sp.STDOUT)
        except sp.CalledProcessError:
            if force_enable:
                raise futils.Fail('Valgrind not found')
            else:
                raise futils.Skip('Valgrind not found')

        try:
            _ = sp.check_output('./valgrind_check/valgrind_check',
                                shell=True, universal_newlines=True,
                                stderr=sp.STDOUT)
        except sp.CalledProcessError:
            raise futils.Skip(
                'Valgrind is required but the Valgrind support has been disabled at compile time')  # noqa: E501

        valgrind_bin = path.join(path.dirname(out), 'valgrind.bin')
        if path.isfile(valgrind_bin):
            return valgrind_bin
        return 'valgrind'

    def add_opt(self, opt):
        """Add option to Valgrind command"""
        self.opts.append(opt)

    def _get_version(self):
        """
        Get Valgrind version represented as integer with patch version ignored
        """
        out = sp.check_output('{} --version'.format(self.valgrind_exe),
                              shell=True, universal_newlines=True)
        version = out.split('valgrind-')[1]
        version_as_int = int(version.rsplit('.', 1)[0].replace('.', ''))
        return version_as_int

    def add_suppression(self, f):
        """
        Add suppression file. Provided file path is
        relative to tests root directory (pmdk/src/test)
        """
        self.opts.append('--suppressions={}'
                         .format(path.join(ROOTDIR, f)))

    def validate_log(self):
        """
        Check Valgrind test result based on Valgrind log file.
        Return True if passed, False otherwise
        """
        if self.tool == NONE:
            return True

        no_ignored = []
        # remove ignored warnings from log file
        with open(self.log_file, 'r+') as f:
            no_ignored = [ln for ln in f if not any(w in ln for w in _IGNORED)]
            f.seek(0)
            f.writelines(no_ignored)
            f.truncate()

        if path.isfile(self.log_file + '.match'):
            # if there is a Valgrind log match file, do nothing - log file
            # will be checked by 'match' tool
            return

        non_zero_errors = 'ERROR SUMMARY: [^0]'
        errors_found = any(re.search(non_zero_errors, ln) for ln in no_ignored)
        if any('Bad pmempool' in ln for ln in no_ignored) or errors_found:
            raise futils.Fail('Valgrind log validation failed')

    def verify(self):
        """
        Check that Valgrind is viable to be used.
        """
        if self.valgrind_exe is None:
            raise futils.Skip('Valgrind not found')

        # verify tool
        cmd = '{} --tool={} --help'.format(self.valgrind_exe, self.tool_name)
        try:
            sp.check_output(cmd, shell=True, stderr=sp.STDOUT)
        except sp.CalledProcessError:
            raise futils.Skip("Valgrind tool '{}' was not found"
                              .format(self.tool_name))


def require_valgrind_enabled(valgrind):
    """
    Enable valgrind tool for given test.

    Used as a test class (tc) decorator.

    Args:
        valgrind (str): valgrind tool

    """

    def wrapped(tc):
        tool = _require_valgrind_common(valgrind)
        ctx.add_requirement(tc, 'enabled_valgrind', tool)

        return tc

    return wrapped


def require_valgrind_disabled(*valgrind):
    """
    Require that the test is not executed with selected valgrind tools.

    Used as a test class (tc) decorator.

    Args:
        *valgrind (*str): variable length arguments list of valgrind
            tools meant to be disabled

    """

    def wrapped(tc):
        disabled_tools = [_require_valgrind_common(v) for v in valgrind]
        ctx.add_requirement(tc, 'disabled_valgrind', disabled_tools)

        return tc

    return wrapped


def _require_valgrind_common(v):
    """
    Validate provided valgrind tool as string
    and translate it to _Tool enum type.

    Args:
        v (str): valgrind tool as string

    Returns:
        valgrind tool as _Tool enum type

    """
    valid_tool_names = [str(t) for t in TOOLS]
    if v not in valid_tool_names:
        sys.exit('used name {} not in valid valgrind tool names which are: {}'
                 .format(v, valid_tool_names))

    str_to_tool = next(t for t in TOOLS if v == str(t))
    return str_to_tool



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2023, Intel Corporation
#


import testframework as t
from testframework import granularity as g


@g.require_granularity(g.ANY)
class Pmem2Config(t.Test):
    test_type = t.Short

    def run(self, ctx):
        filepath = ctx.create_holey_file(16 * t.MiB, 'testfile1')
        ctx.exec('pmem2_config', self.test_case, filepath)


@g.no_testdir()
class Pmem2ConfigNoDir(t.Test):
    test_type = t.Short

    def run(self, ctx):
        ctx.exec('pmem2_config', self.test_case)


class TEST0(Pmem2ConfigNoDir):
    """allocation and dealocation of pmem2_config"""
    test_case = "test_cfg_create_and_delete_valid"


class TEST1(Pmem2ConfigNoDir):
    """allocation of pmem2_config in case of missing memory in system"""
    labels = ['fault_injection']
    test_case = "test_alloc_cfg_enomem"


class TEST2(Pmem2ConfigNoDir):
    """deleting null pmem2_config"""
    test_case = "test_delete_null_config"


class TEST3(Pmem2ConfigNoDir):
    """set valid granularity in the config"""
    test_case = "test_config_set_granularity_valid"


class TEST4(Pmem2ConfigNoDir):
    """set invalid granularity in the config"""
    test_case = "test_config_set_granularity_invalid"


class TEST5(Pmem2ConfigNoDir):
    """setting offset which is too large"""
    test_case = "test_set_offset_too_large"


class TEST6(Pmem2ConfigNoDir):
    """setting a valid offset"""
    test_case = "test_set_offset_success"


class TEST7(Pmem2ConfigNoDir):
    """setting a valid length"""
    test_case = "test_set_length_success"


class TEST8(Pmem2ConfigNoDir):
    """setting maximum possible offset"""
    test_case = "test_set_offset_max"


class TEST9(Pmem2ConfigNoDir):
    """setting a valid sharing"""
    test_case = "test_set_sharing_valid"


class TEST10(Pmem2ConfigNoDir):
    """setting a invalid sharing"""
    test_case = "test_set_sharing_invalid"


class TEST11(Pmem2ConfigNoDir):
    """
    setting a valid protection flags
    """
    test_case = "test_set_valid_prot_flag"


class TEST12(Pmem2ConfigNoDir):
    """
    setting a invalid protection flags
    """
    test_case = "test_set_invalid_prot_flag"



#!../env.py
#
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2020-2023, Intel Corporation

import testframework as t
from testframework import granularity as g
import futils
import os


@t.require_valgrind_enabled('pmemcheck')
# XXX In the match file, there are two possible numbers of errors. It varies
# from compiler to compiler. There should be only one number when pmemcheck
# will be fixed. Please also remove the below requirement after pmemcheck fix.
# https://github.com/pmem/valgrind/pull/76
@g.require_granularity(g.CL_OR_LESS)
class PMEM2_PERSIST(t.Test):
    test_type = t.Medium
    available_granularity = None

    def run(self, ctx):
        filepath = ctx.create_holey_file(2 * t.MiB, 'testfile')
        ctx.exec('pmem2_persist_valgrind', self.test_case, filepath)


class TEST0(PMEM2_PERSIST):
    """persist continuous data in a range of pmem"""
    test_case = "test_persist_continuous_range"


class TEST1(PMEM2_PERSIST):
    """persist discontinuous data in a range of pmem"""
    test_case = "test_persist_discontinuous_range"


class TEST2(PMEM2_PERSIST):
    """persist part of discontinuous data in a range of pmem"""
    test_case = "test_persist_discontinuous_range_partially"

    def run(self, ctx):
        filepath = ctx.create_holey_file(16 * t.KiB, 'testfile')
        ctx.exec('pmem2_persist_valgrind', self.test_case, filepath)
        pmemecheck_log = os.path.join(
            os.getcwd(), 'pmem2_persist_valgrind', 'pmemcheck2.log')
        futils.tail(pmemecheck_log, 2)


class TEST3(PMEM2_PERSIST):
    """persist data in a range of the memory mapped by mmap()"""
    test_case = "test_persist_nonpmem_data"



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2020-2023, Intel Corporation
#

import testframework as t
from testframework import granularity as g


@g.no_testdir()
class TEST0(t.Test):
    test_type = t.Medium
    labels = ['fault_injection']

    def run(self, ctx):
        ctx.exec('obj_critnib')



# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2023, Intel Corporation
#
"""Test type context classes"""

import context as ctx


class _TestType(metaclass=ctx.CtxType):
    """Base class for a test duration."""


class Short(_TestType):
    pass


class Medium(_TestType):
    pass


class Long(_TestType):
    pass


class Check(_TestType):
    includes = [Short, Medium]


class All(_TestType):
    includes = [Short, Medium, Long]



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2020-2022, Intel Corporation
#

import testframework as t


class MovntAlignCommon(t.Test):
    test_type = t.Short
    filesize = 4 * t.MiB

    def run_cases(self, ctx):
        ctx.exec('pmem2_movnt_align', self.filepath, "C")
        ctx.exec('pmem2_movnt_align', self.filepath, "F")
        ctx.exec('pmem2_movnt_align', self.filepath, "B")
        ctx.exec('pmem2_movnt_align', self.filepath, "S")

    def run(self, ctx):
        self.filepath = ctx.create_holey_file(self.filesize, 'testfile',)
        self.run_cases(ctx)


class Pmem2MovntAlign(MovntAlignCommon):
    threshold = None
    threshold_values = ['0', '99999']
    envs0 = ()

    def run(self, ctx):
        for env in self.envs0:
            ctx.env[env] = '0'

        super().run(ctx)
        for tv in self.threshold_values:
            ctx.env['PMEM_MOVNT_THRESHOLD'] = tv
            self.run_cases(ctx)


@t.require_valgrind_enabled('pmemcheck')
class MovntAlignCommonValgrind(Pmem2MovntAlign):
    test_type = t.Medium

    def run(self, ctx):
        ctx.env['VALGRIND_OPTS'] = "--mult-stores=yes"
        super().run(ctx)


class TEST0(Pmem2MovntAlign):
    pass


@t.require_architectures('x86_64')
class TEST1(Pmem2MovntAlign):
    envs0 = ("PMEM_AVX512F",)


@t.require_architectures('x86_64')
class TEST2(Pmem2MovntAlign):
    envs0 = ("PMEM_AVX512F", "PMEM_AVX",)


class TEST3(MovntAlignCommon):
    def run(self, ctx):
        ctx.env['PMEM_NO_MOVNT'] = '1'
        super().run(ctx)


class TEST4(MovntAlignCommon):
    def run(self, ctx):
        ctx.env['PMEM_NO_MOVNT'] = '1'
        ctx.env['PMEM_NO_GENERIC_MEMCPY'] = '1'
        super().run(ctx)


class TEST5(MovntAlignCommonValgrind):
    pass


@t.require_architectures('x86_64')
class TEST6(MovntAlignCommonValgrind):
    envs0 = ("PMEM_AVX512F",)


@t.require_architectures('x86_64')
class TEST7(MovntAlignCommonValgrind):
    envs0 = ("PMEM_AVX512F", "PMEM_AVX",)


class TEST8(MovntAlignCommonValgrind):
    def run(self, ctx):
        ctx.env['PMEM_NO_MOVNT'] = '1'
        super().run(ctx)


class TEST9(MovntAlignCommonValgrind):
    def run(self, ctx):
        ctx.env['PMEM_NO_MOVNT'] = '1'
        ctx.env['PMEM_NO_GENERIC_MEMCPY'] = '1'
        super().run(ctx)


@t.require_architectures('x86_64')
class TEST10(Pmem2MovntAlign):
    envs1 = ("PMEM_MOVDIR64B",)



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2020, Intel Corporation
#

import testframework as t
from testframework import granularity as g


@g.require_granularity(g.ANY)
class Pmem2Perror(t.Test):
    test_type = t.Short

    def run(self, ctx):
        ctx.exec('pmem2_perror', self.test_case, stderr_file=self.stderr_file)


class TEST0(Pmem2Perror):
    test_case = 'test_fail_pmem2_func_simple'
    stderr_file = 'test0.log'


class TEST1(Pmem2Perror):
    test_case = 'test_fail_pmem2_func_format'
    stderr_file = 'test1.log'


class TEST2(Pmem2Perror):
    test_case = 'test_fail_system_func_simple'
    stderr_file = 'test2.log'


class TEST3(Pmem2Perror):
    test_case = 'test_fail_system_func_format'
    stderr_file = 'test3.log'


class TEST4(Pmem2Perror):
    test_case = 'test_fail_pmem2_syscall_simple'
    stderr_file = 'test4.log'


class TEST5(Pmem2Perror):
    test_case = 'test_fail_pmem2_syscall_format'
    stderr_file = 'test5.log'


class TEST6(Pmem2Perror):
    test_case = 'test_simple_err_to_errno_check'
    stderr_file = None



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2020-2024, Intel Corporation
#

import testframework as t
from testframework import granularity as g
from testframework import tools
from testframework import futils


NO_FLAGS = 0
PMEM_F_MEM_NONTEMPORAL = 1
PMEM_F_MEM_TEMPORAL = 2
PMEM_F_MEM_NONTEMPORAL_v_PMEM_F_MEM_TEMPORAL = 3
PMEM_F_MEM_WC = 4
PMEM_F_MEM_WB = 5
PMEM_F_MEM_NOFLUSH = 6
'''
ALL_FLAGS = PMEM_F_MEM_NODRAIN | PMEM_F_MEM_NOFLUSH |
            PMEM_F_MEM_NONTEMPORAL | PMEM_F_MEM_TEMPORAL |
            PMEM_F_MEM_WC | PMEM_F_MEM_WB
'''
ALL_FLAGS = 7

# This match is valid for all cases with a BYTE
# granularity and small (<256) data size.
MATCH_BYTE_SMALL = \
    (
        (NO_FLAGS,                                      128, "t"),
        (PMEM_F_MEM_NONTEMPORAL,                        128, "nt"),
        (PMEM_F_MEM_TEMPORAL,                           128, "t"),
        (PMEM_F_MEM_NONTEMPORAL_v_PMEM_F_MEM_TEMPORAL,  128, "nt"),
        (PMEM_F_MEM_WC,                                 128, "t"),
        (PMEM_F_MEM_WB,                                 128, "t"),
        (PMEM_F_MEM_NOFLUSH,                            128, "t"),
        (ALL_FLAGS,                                     128, "t")
    )

# This match is valid for all cases with a BYTE
# granularity and big (>256) data size.
MATCH_BYTE_BIG = \
    (
        (NO_FLAGS,                                      1024, "t"),
        (PMEM_F_MEM_NONTEMPORAL,                        1024, "nt"),
        (PMEM_F_MEM_TEMPORAL,                           1024, "t"),
        (PMEM_F_MEM_NONTEMPORAL_v_PMEM_F_MEM_TEMPORAL,  1024, "nt"),
        (PMEM_F_MEM_WC,                                 1024, "t"),
        (PMEM_F_MEM_WB,                                 1024, "t"),
        (PMEM_F_MEM_NOFLUSH,                            1024, "t"),
        (ALL_FLAGS,                                     1024, "t")
    )

# This match is valid for all cases with a PAGE/CACHELINE
# granularity and small (<256) data size.
MATCH_PAGE_CACHELINE_SMALL = \
    (
        (NO_FLAGS,                                      128, "t"),
        (PMEM_F_MEM_NONTEMPORAL,                        128, "nt"),
        (PMEM_F_MEM_TEMPORAL,                           128, "t"),
        (PMEM_F_MEM_NONTEMPORAL_v_PMEM_F_MEM_TEMPORAL,  128, "nt"),
        (PMEM_F_MEM_WC,                                 128, "nt"),
        (PMEM_F_MEM_WB,                                 128, "t"),
        (PMEM_F_MEM_NOFLUSH,                            128, "t"),
        (ALL_FLAGS,                                     128, "t")
    )

# This match is valid for all cases with a PAGE/CACHELINE
# granularity and big (>256) data size.
MATCH_PAGE_CACHELINE_BIG = \
    (
        (NO_FLAGS,                                      1024, "nt"),
        (PMEM_F_MEM_NONTEMPORAL,                        1024, "nt"),
        (PMEM_F_MEM_TEMPORAL,                           1024, "t"),
        (PMEM_F_MEM_NONTEMPORAL_v_PMEM_F_MEM_TEMPORAL,  1024, "nt"),
        (PMEM_F_MEM_WC,                                 1024, "nt"),
        (PMEM_F_MEM_WB,                                 1024, "t"),
        (PMEM_F_MEM_NOFLUSH,                            1024, "t"),
        (ALL_FLAGS,                                     1024, "t")
    )

SSE2 = 1
AVX = 2
AVX512 = 3
MOVDIR64B = 4

VARIANT_LIBC = 'libc'
VARIANT_GENERIC = 'generic'
VARIANT_SSE2 = 'sse2'
VARIANT_AVX = 'avx'
VARIANT_AVX512F = 'avx512f'
VARIANT_MOVDIR64B = 'movdir64b'


@t.require_build('debug')
@t.require_architectures('x86_64')
class Pmem2MemExt(t.Test):
    test_type = t.Short
    filesize = 4 * t.MiB

    available_arch = SSE2
    variant = VARIANT_SSE2

    # By default data size is 128 - this is smaller than threshold value (256)
    # to predict usage of temporal stores. This value is overridden in some
    # tests to values bigger than 256.
    data_size = 128

    oper = ("C", "M", "S")

    def setup(self, ctx):
        super().setup(ctx)
        ret = tools.Tools(ctx.env, ctx.build).cpufd()
        self.check_arch(ctx.variant(), ret.returncode)

        # XXX all tests with VARIANT_AVX512F or VARIANT_MOVDIR64B
        # are disabled under Valgrind until the issue
        # https://github.com/pmem/pmdk/issues/5640 and
        # https://github.com/pmem/pmdk/issues/6129 are fixed.
        if ctx.valgrind is not None and ctx.valgrind.tool.name != "NONE":
            if ctx.variant() == VARIANT_AVX512F:
                raise futils.Skip("AVX512F unavailable under Valigrind")
            if ctx.variant() == VARIANT_MOVDIR64B:
                raise futils.Skip("MOVDIR64B unavailable under Valigrind")

    def check_arch(self, variant, available_arch):
        if variant == VARIANT_MOVDIR64B:
            if available_arch < MOVDIR64B:
                raise futils.Skip("MOVDIR64B unavailable")

            is_movdir64b_enabled = tools.envconfig['PMEM2_MOVDIR64B_ENABLED']
            if is_movdir64b_enabled == "0":
                raise futils.Skip("MOVDIR64B disabled at build time")

        if variant == VARIANT_AVX512F:
            if available_arch < AVX512:
                raise futils.Skip("AVX512F unavailable")

            is_avx512f_enabled = tools.envconfig['PMEM2_AVX512F_ENABLED']
            if is_avx512f_enabled == "0":
                raise futils.Skip("AVX512F disabled at build time")

        if variant == VARIANT_AVX and available_arch < AVX:
            raise futils.Skip("AVX unavailable")

    def check_log(self, ctx, match, type, flag):
        with open(self.log_files['pmem2'], 'r') as f:
            str_val = f.read()

            # count function match, only one log should occur at the time
            count = str_val.count(match)
            if count != 1:
                raise futils.Fail(
                    "Pattern: {} occurs {} times. One expected. "
                    "Type: {} Flag id: {}"
                    .format(match, count, type, flag))

    def create_match(self, variant, oper, store_type):
        match = ""
        if variant == VARIANT_LIBC:
            if oper == "C" or oper == "M":
                match = "memmove_nodrain_libc"
            elif oper == "S":
                match = "memset_nodrain_libc"
            return match

        if variant == VARIANT_GENERIC:
            if oper == "C" or oper == "M":
                match = "memmove_nodrain_generic"
            elif oper == "S":
                match = "memset_nodrain_generic"
            return match

        if oper in ("C", "M"):
            match += "memmove_mov"
        elif oper == "S":
            match += "memset_mov"
        else:
            raise futils.Fail(
                "Operation: {} not supported.".format(oper))

        if store_type == "nt":
            match += store_type

        if variant == VARIANT_SSE2:
            match += "_sse2"
        elif variant == VARIANT_AVX:
            match += "_avx"
        elif variant == VARIANT_AVX512F:
            match += "_avx512f"
        elif variant == VARIANT_MOVDIR64B:
            match += "_movdir64b"
        else:
            assert False

        return match

    def run(self, ctx):
        ctx.env['PMEM2_LOG_LEVEL'] = '15'

        if ctx.wc_workaround() == 'on':
            ctx.env['PMEM_WC_WORKAROUND'] = '1'
        elif ctx.wc_workaround() == 'off':
            ctx.env['PMEM_WC_WORKAROUND'] = '0'

        if ctx.variant() == VARIANT_LIBC:
            ctx.env['PMEM_NO_MOVNT'] = '1'
            ctx.env['PMEM_NO_GENERIC_MEMCPY'] = '1'
        elif ctx.variant() == VARIANT_GENERIC:
            ctx.env['PMEM_NO_MOVNT'] = '1'
        elif ctx.variant() == VARIANT_SSE2:
            ctx.env['PMEM_AVX'] = '0'
            ctx.env['PMEM_AVX512F'] = '0'
            ctx.env['PMEM_MOVDIR64B'] = '0'
        elif ctx.variant() == VARIANT_AVX:
            ctx.env['PMEM_AVX'] = '1'
            ctx.env['PMEM_AVX512F'] = '0'
            ctx.env['PMEM_MOVDIR64B'] = '0'
        elif ctx.variant() == VARIANT_AVX512F:
            ctx.env['PMEM_AVX'] = '0'
            ctx.env['PMEM_AVX512F'] = '1'
            ctx.env['PMEM_MOVDIR64B'] = '0'
        elif ctx.variant() == VARIANT_MOVDIR64B:
            ctx.env['PMEM_AVX'] = '0'
            ctx.env['PMEM_AVX512F'] = '0'
            ctx.env['PMEM_MOVDIR64B'] = '1'

        filepath = ctx.create_holey_file(self.filesize, 'testfile',)

        for tc in self.test_case:
            for o in self.oper:
                flag_id = tc[0]
                size = tc[1]
                store_type = tc[2]
                match = self.create_match(ctx.variant(), o, store_type)
                ctx.exec('pmem2_mem_ext', filepath, o, size, flag_id)
                self.check_log(ctx, match, o, flag_id)


@t.add_params('variant', [VARIANT_LIBC, VARIANT_GENERIC])
@t.add_params('wc_workaround', ['default'])
class TEST0(Pmem2MemExt):
    test_case = [(NO_FLAGS, 1024, "")]


@g.require_granularity(g.PAGE, g.CACHELINE)
@t.add_params('variant', [VARIANT_SSE2, VARIANT_AVX, VARIANT_AVX512F])
@t.add_params('wc_workaround', ['on', 'off', 'default'])
class TEST1(Pmem2MemExt):
    test_case = MATCH_PAGE_CACHELINE_SMALL


@g.require_granularity(g.BYTE)
@t.add_params('variant', [VARIANT_SSE2, VARIANT_AVX, VARIANT_AVX512F])
@t.add_params('wc_workaround', ['on', 'off', 'default'])
class TEST2(Pmem2MemExt):
    test_case = MATCH_BYTE_SMALL


@g.require_granularity(g.PAGE, g.CACHELINE)
@t.add_params('variant', [VARIANT_SSE2, VARIANT_AVX, VARIANT_AVX512F])
@t.add_params('wc_workaround', ['on', 'off', 'default'])
class TEST3(Pmem2MemExt):
    test_case = MATCH_PAGE_CACHELINE_BIG


@g.require_granularity(g.BYTE)
@t.add_params('variant', [VARIANT_SSE2, VARIANT_AVX, VARIANT_AVX512F])
@t.add_params('wc_workaround', ['on', 'off', 'default'])
class TEST4(Pmem2MemExt):
    test_case = MATCH_BYTE_BIG


@t.add_params('variant', [VARIANT_MOVDIR64B])
@t.add_params('wc_workaround', ['on', 'off', 'default'])
class TEST5(Pmem2MemExt):
    test_case = [(PMEM_F_MEM_NONTEMPORAL, 128, "nt")]


@t.add_params('variant', [VARIANT_MOVDIR64B])
@t.add_params('wc_workaround', ['on', 'off', 'default'])
class TEST6(Pmem2MemExt):
    test_case = [(PMEM_F_MEM_NONTEMPORAL, 1024, "nt")]
